from pathlib import Path

from ..dataset import MLDataset
from ..dataset.features import get_features_from_raw_features
from ..dataset.utils import df_to_numpy
from ..icarus import IcarusPrediction, IcarusRSM
from ..icarus.analyse import analyse_icarus_predictions


def generate_perturbed_predictions(
    model: IcarusRSM,
    dataset: MLDataset,
    rng,  #: np.random.Generator,
    n_samples: int,
    prediction_path: Path,
    overwrite_rsm_prediction: bool,
    perturbation,  #: Callable[[pd.DataFrame], pd.DataFrame],
    progress=None,
    **kwargs,
):  # -> pd.DataFrame:
    import joblib
    import numpy as np
    import pandas as pd

    if prediction_path.exists() and not overwrite_rsm_prediction:
        # FIXME: support saving to NetCDF files instead
        return joblib.load(prediction_path)

    if progress is not None:
        progress.update_major(
            value=0,
            min=0,
            max=3 + n_samples,
            format="Evaluating on the Test Dataset",
        )

    X_raw = perturbation(dataset.X_raw.copy(deep=True))

    if progress is not None:
        progress.update_major(format="Generating the Perturbed Dataset")

    X_prtb = dataset.X_scaler.transform(
        get_features_from_raw_features(X_raw, progress=progress)
    )
    Y_base = dataset.Y_raw

    prtb_predictions = []
    for i in range(n_samples):
        if progress is not None:
            progress.update_major(
                format=f"Predicting on the Perturbed Dataset {i}/{n_samples}"
            )

        prtb_predictions.append(
            model.predict(X_prtb, rng, progress=progress, **kwargs)
        )

    if progress is not None:
        progress.update_major(
            format="Combining the Predictions on the Perturbed Dataset"
        )

        progress.update_minor(
            value=0,
            min=0,
            max=len(X_prtb),
            format="Perturbed Prediction %v/%m",
        )

    combined_prtb_predictions = IcarusPrediction(
        prediction=[],
        uncertainty=[],
        confidence=[],
    )

    for i in range(len(X_prtb)):
        predictions = np.array([p.prediction[i] for p in prtb_predictions])
        uncertainties = np.array([p.uncertainty[i] for p in prtb_predictions])
        confidences = np.array([p.confidence[i] for p in prtb_predictions])

        def combine_predictions(Y_pred, I_pred, rng, **kwargs):
            return (
                np.mean(Y_pred)
                if len(Y_pred) > 0
                else np.mean(
                    prtb_predictions[
                        rng.choice(len(prtb_predictions))
                    ].prediction[i]
                )
            )

        cp = analyse_icarus_predictions(
            IcarusPrediction(
                prediction=predictions,
                uncertainty=uncertainties,
                confidence=confidences,
            ),
            combine_predictions,
            rng,
            n_uncertain_samples=1,
            n_analysis_runs=10,
            progress=None,
        )

        combined_prtb_predictions.prediction.append(cp.prediction)
        combined_prtb_predictions.uncertainty.append(cp.uncertainty)
        combined_prtb_predictions.confidence.append(cp.confidence)

        if progress is not None:
            progress.update_minor()

    prtb_predictions = IcarusPrediction(
        prediction=np.array(combined_prtb_predictions.prediction).reshape(
            -1, 1
        ),
        uncertainty=np.array(combined_prtb_predictions.uncertainty).reshape(
            -1, 1
        ),
        confidence=np.array(combined_prtb_predictions.confidence),
    )

    if progress is not None:
        progress.update_major(format="Assembling the Perturbation Prediction")
        progress.update_minor(value=0, format="")

    Y_pred = np.concatenate(
        [
            Y_base.index.get_level_values(0)
            .to_numpy()
            .reshape(
                (
                    Y_base.index.levels[0].size,
                    Y_base.index.levels[1].size,
                    1,
                )
            ),
            Y_base.index.get_level_values(1)
            .to_numpy()
            .reshape(
                (
                    Y_base.index.levels[0].size,
                    Y_base.index.levels[1].size,
                    1,
                )
            ),
            df_to_numpy(Y_base).reshape(
                (Y_base.index.levels[0].size, Y_base.index.levels[1].size, 1)
            ),
            dataset.Y_scaler.inverse_transform(
                prtb_predictions.prediction
            ).reshape(
                (Y_base.index.levels[0].size, Y_base.index.levels[1].size, 1)
            ),
            (prtb_predictions.uncertainty * dataset.Y_scaler.scale_).reshape(
                (Y_base.index.levels[0].size, Y_base.index.levels[1].size, 1)
            ),
            prtb_predictions.confidence.reshape(
                (Y_base.index.levels[0].size, Y_base.index.levels[1].size, 1)
            ),
        ],
        axis=2,
    )

    df = pd.DataFrame(
        Y_pred.reshape(
            Y_pred.shape[0] * Y_pred.shape[1],
            Y_pred.shape[2],
        ),
        columns=[
            "time",
            "level",
            "log10_ccn_baseline",
            "log10_ccn_perturbed_pred",
            "log10_ccn_perturbed_stdv",
            "log10_ccn_perturbed_conf",
        ],
    ).set_index(["time", "level"])

    joblib.dump(df, prediction_path)

    return df



from ...labels.time import get_output_time
from ...netcdf import TrajectoryDatasets


def interpolate_anthropogenic_emissions(ds: TrajectoryDatasets, key: str):
    import scipy as sp

    out_t = get_output_time(ds)
    out_h = ds.out["lev"][:].data

    ant_t = ds.ant["time"][:].data
    ant_h = ds.ant["mid_layer_height"][:].data

    ant_t_h = ds.ant[key][:].T

    ant_t_h_int = sp.interpolate.interp2d(
        x=ant_h,
        y=ant_t,
        z=ant_t_h,
        kind="linear",
        bounds_error=False,
        fill_value=0.0,
    )

    return ant_t_h_int(x=out_h, y=out_t)


def get_ant_emissions_features(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    return pd.DataFrame(
        {
            "time": np.repeat(get_output_time(ds), ds.out["lev"].shape[0]),
            "level": np.tile(ds.out["lev"][:].data, ds.out["time"].shape[0]),
            "ant_co": interpolate_anthropogenic_emissions(ds, "co").flatten(),
            "ant_nox": interpolate_anthropogenic_emissions(
                ds, "nox"
            ).flatten(),
            "ant_co2": interpolate_anthropogenic_emissions(
                ds, "co2"
            ).flatten(),
            "ant_nh3": interpolate_anthropogenic_emissions(
                ds, "nh3"
            ).flatten(),
            "ant_ch4": interpolate_anthropogenic_emissions(
                ds, "ch4"
            ).flatten(),
            "ant_so2": interpolate_anthropogenic_emissions(
                ds, "so2"
            ).flatten(),
            "ant_nmvoc": interpolate_anthropogenic_emissions(
                ds, "nmvoc"
            ).flatten(),
            "ant_alcohols": interpolate_anthropogenic_emissions(
                ds, "alcohols"
            ).flatten(),
            "ant_ethane": interpolate_anthropogenic_emissions(
                ds, "ethane"
            ).flatten(),
            "ant_propane": interpolate_anthropogenic_emissions(
                ds, "propane"
            ).flatten(),
            "ant_butanes": interpolate_anthropogenic_emissions(
                ds, "butanes"
            ).flatten(),
            "ant_pentanes": interpolate_anthropogenic_emissions(
                ds, "pentanes"
            ).flatten(),
            "ant_hexanes": interpolate_anthropogenic_emissions(
                ds, "hexanes"
            ).flatten(),
            "ant_ethene": interpolate_anthropogenic_emissions(
                ds, "ethene"
            ).flatten(),
            "ant_propene": interpolate_anthropogenic_emissions(
                ds, "propene"
            ).flatten(),
            "ant_acetylene": interpolate_anthropogenic_emissions(
                ds, "acetylene"
            ).flatten(),
            "ant_isoprene": interpolate_anthropogenic_emissions(
                ds, "isoprene"
            ).flatten(),
            "ant_monoterpenes": interpolate_anthropogenic_emissions(
                ds, "monoterpenes"
            ).flatten(),
            "ant_other_alkenes_and_alkynes": (
                interpolate_anthropogenic_emissions(
                    ds, "other-alkenes-and-alkynes"
                ).flatten()
            ),
            "ant_benzene": interpolate_anthropogenic_emissions(
                ds, "benzene"
            ).flatten(),
            "ant_toluene": interpolate_anthropogenic_emissions(
                ds, "toluene"
            ).flatten(),
            "ant_xylene": interpolate_anthropogenic_emissions(
                ds, "xylene"
            ).flatten(),
            "ant_trimethylbenzene": interpolate_anthropogenic_emissions(
                ds, "trimethylbenzene"
            ).flatten(),
            "ant_other_aromatics": interpolate_anthropogenic_emissions(
                ds, "other-aromatics"
            ).flatten(),
            "ant_esters": interpolate_anthropogenic_emissions(
                ds, "esters"
            ).flatten(),
            "ant_ethers": interpolate_anthropogenic_emissions(
                ds, "ethers"
            ).flatten(),
            "ant_formaldehyde": interpolate_anthropogenic_emissions(
                ds, "formaldehyde"
            ).flatten(),
            "ant_other_aldehydes": interpolate_anthropogenic_emissions(
                ds, "other-aldehydes"
            ).flatten(),
            "ant_total_ketones": interpolate_anthropogenic_emissions(
                ds, "total-ketones"
            ).flatten(),
            "ant_total_acids": interpolate_anthropogenic_emissions(
                ds, "total-acids"
            ).flatten(),
            "ant_other_vocs": interpolate_anthropogenic_emissions(
                ds, "other-VOCs"
            ).flatten(),
        }
    ).set_index(["time", "level"])



from PyQt5 import QtCore, QtGui, QtWidgets

from ..layouts import about
from ..resources import resource_path


class About(QtWidgets.QDialog):
    def __init__(self, gui):
        super(About, self).__init__()

        self.ab = about.Ui_Dialog()
        self.ab.setupUi(self)
        self.ab.okgreat.clicked.connect(self.reject)
        self.setWindowIcon(QtGui.QIcon(resource_path("icons/sosaa-icon.svg")))
        self.ab.logo.setPixmap(
            QtGui.QPixmap(
                resource_path(
                    "icons/sosaa-splash-dark.svg"
                    if gui.dark
                    else "icons/sosaa-splash-light.svg"
                )
            )
        )


def init_gui_help(gui):
    def openAboutHelp():
        About(gui).exec()

    gui.actionAbout_SOSAA.triggered.connect(openAboutHelp)

    def openSOSAAWebpage():
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl(
                "https://www.helsinki.fi/en/researchgroups/"
                "multi-scale-modelling/sosaa"
            )
        )

    gui.actionSOSAA_webpage.triggered.connect(openSOSAAWebpage)

    def openSOSAAManual():
        return

    gui.actionOnline_manual.triggered.connect(openSOSAAManual)



from pathlib import Path

from ..browse import browsePath
from ..syntax import PythonHighlighter
from .build import build_sosaa_rsm
from .plot import update_rsm_plots
from .predict import predict_sosaa_rsm
from .progress import RsmMajorMinorProgress


def init_rsm_gui(gui):
    gui.rsm_dataset = None
    gui.rsm_model = None
    gui.rsm_prediction = None
    gui.rsm_plots_dirty = True

    # Initialise the progress bar for RSM training and prediction
    gui.rsm_build_progress = RsmMajorMinorProgress(
        gui.rsm_build_progress_major,
        gui.rsm_build_progress_minor,
    )
    gui.rsm_build_progress.update_major(
        value=0, format="No SOSAA RSM is loaded"
    )
    gui.rsm_build_progress.update_minor(value=0, format="")

    gui.rsm_predict_progress = RsmMajorMinorProgress(
        gui.rsm_predict_progress_major,
        gui.rsm_predict_progress_minor,
    )
    gui.rsm_predict_progress.update_major(value=0, format="")
    gui.rsm_predict_progress.update_minor(value=0, format="")

    # Connect the RSM training and evaluation buttons
    gui.rsm_build.clicked.connect(
        lambda: build_sosaa_rsm(gui, rsm_should_exist=False)
    )
    gui.rsm_load.clicked.connect(
        lambda: build_sosaa_rsm(gui, rsm_should_exist=True)
    )

    # Connect and disable the RSM prediction button
    gui.rsm_predict.clicked.connect(lambda: predict_sosaa_rsm(gui))
    gui.rsm_predict.setEnabled(False)

    # Initialise the RSM configuration inputs
    gui.browse_rsm.clicked.connect(lambda: _change_rsm_file(gui))
    gui.browse_rsm_output.clicked.connect(
        lambda: _change_rsm_prediction_file(gui)
    )

    gui.rsm_perturbation.setPlaceholderText("return inputs")
    gui.rsm_perturbation_highlight = PythonHighlighter(
        gui,
        gui.rsm_perturbation.document(),
    )

    # Listen to RSM tab changes to lazily update the plots
    gui.rsm_subtab.currentChanged.connect(lambda i: _tab_switched(gui, i))


def _change_rsm_file(gui):
    path = browsePath(title="Choose the RSM file")

    if path is None:
        return

    if str(path).startswith(str(Path.cwd())):
        path = f"./{path.relative_to(Path.cwd())}"

    gui.rsm_path.setText(str(path))


def _change_rsm_prediction_file(gui):
    path = browsePath(title="Choose the RSM prediction output file")

    if path is None:
        return

    if str(path).startswith(str(Path.cwd())):
        path = f"./{path.relative_to(Path.cwd())}"

    gui.rsm_output.setText(str(path))


def _tab_switched(gui, _i):
    # Fake-access the i variable
    _i = _i

    if gui.rsm_subtab.currentWidget() == gui.rsm_prediction_tab:
        update_rsm_plots(gui)



from pathlib import Path

from PyQt5 import QtCore, QtWidgets


def update_gui_from_settings(settings, gui, raw):
    _update_gui_from_main_settings(settings, gui)
    _update_gui_from_flag_settings(settings, gui)
    _update_gui_from_time_settings(settings, gui)
    _update_gui_from_output_settings(settings, gui)
    _update_gui_from_custom_settings(settings, gui)
    _update_gui_from_compile_settings(settings, gui)
    _update_gui_from_run_settings(settings, gui)
    _update_gui_from_rsm_settings(settings, gui)
    _update_gui_from_raw_settings(gui, raw)


def _update_gui_from_main_settings(settings, gui):
    main = settings.get("nml_main", dict())
    guis = settings.get("nml_gui", dict())

    # Pretty-print the main dir if relative to the working dir
    main_dir = str(Path(main.get("work_dir", "./sosaa")).resolve())
    if main_dir.startswith(str(Path.cwd())):
        main_dir = f"./{Path(main_dir).relative_to(Path.cwd())}"
    gui.main_dir.setText(main_dir)

    main_dir = Path(main.get("work_dir", "./sosaa")).resolve()

    # Pretty-print the code dir if relative to the main dir
    code_dir = main.get("code_dir", "src")
    if code_dir.startswith(str(main_dir)):
        code_dir = f"{Path(code_dir).relative_to(main_dir)}"
    gui.code_dir.setText(code_dir)

    # Pretty-print the case dir if relative to the main dir
    case_dir = main.get("case_dir", "casedir")
    if case_dir.startswith(str(main_dir)):
        case_dir = f"{Path(case_dir).relative_to(main_dir)}"
    gui.case_dir.setText(case_dir)

    gui.casename_dir.setText(guis.get("main_casename_dir", "sample_aerosol"))

    # Split the chem_dir into the outer dir and the chemistry name
    chem_dir = main.get("chem_dir", None)
    chemall_dir = guis.get("main_chemall_dir", None)
    chemname_dir = guis.get("main_chemname_dir", None)

    if chemall_dir is None or chemname_dir is None:
        if chem_dir is not None:
            chemall_dir = str(Path(chem_dir).parent)
            chemname_dir = Path(chem_dir).name
        else:
            chemall_dir = "chemistry"
            chemname_dir = "sample_aerosol"

    # Pretty-print the chem dir if relative to the main dir
    if chemall_dir.startswith(str(main_dir)):
        chemall_dir = f"{Path(chemall_dir).relative_to(main_dir)}"

    gui.chem_dir.setText(chemall_dir)
    gui.chemname_dir.setText(chemname_dir)

    # Pretty-print the input dir if relative to the working dir
    input_dir = str(Path(main.get("input_dir", "")).resolve())
    if input_dir.startswith(str(Path.cwd())):
        input_dir = f"./{Path(input_dir).relative_to(Path.cwd())}"
    gui.input_dir.setText(input_dir)

    # Pretty-print the output dir if relative to the working dir
    output_dir = str(Path(main.get("output_dir", "")).resolve())
    if output_dir.startswith(str(Path.cwd())):
        output_dir = f"./{Path(output_dir).relative_to(Path.cwd())}"
    gui.output_dir.setText(output_dir)
    gui.currentOutputPath.setText(output_dir)


def _update_gui_from_flag_settings(settings, gui):
    flag = settings.get("nml_flag", dict())
    aer = settings.get("aer_flag", dict()).get("options", dict())

    # flag_emis affects flag_emis, dt_emis, and output_list_emi
    flag_emis = flag.get("flag_emis", 2) != 0
    gui.flag_emis.setChecked(flag_emis)
    gui.dt_emis.setEnabled(flag_emis)
    gui.label_dt_emis.setEnabled(flag_emis)
    gui.group_output_list_emi.setEnabled(flag_emis)

    # Note: Soil emissions are currently always disabled
    gui.flag_emis_soil.setChecked(False)

    # flag_chem affects flag_chem, dt_chem, and chem_dir
    flag_chem = flag.get("flag_chem", 1) != 0
    gui.flag_chem.setChecked(flag_chem)
    gui.dt_chem.setEnabled(flag_chem)
    gui.label_dt_chem.setEnabled(flag_chem)
    gui.label_chem_dir.setEnabled(flag_chem)
    gui.chem_dir.setEnabled(flag_chem)
    gui.browse_chem.setEnabled(flag_chem)

    flag_mix_chem = flag.get("flag_mix_chem", 1) != 0
    gui.flag_mix_chem.setChecked(flag_mix_chem)

    # flag_gasdrydep affects flag_gasdrydep, dt_depo, and output_list_Vd
    flag_gasdrydep = flag.get("flag_gasdrydep", 0) != 0
    gui.flag_gasdrydep.setChecked(flag_gasdrydep)
    gui.dt_depo.setEnabled(flag_gasdrydep)
    gui.label_dt_depo.setEnabled(flag_gasdrydep)
    gui.group_output_list_Vd.setEnabled(flag_gasdrydep)

    # flag_aero affects flag_aero, dt_aero, dt_uhma, and aero_start_date
    flag_aero = flag.get("flag_aero", 1) != 0
    gui.flag_aero.setChecked(flag_aero)
    gui.dt_aero.setEnabled(flag_aero)
    gui.label_dt_aero.setEnabled(flag_aero)
    gui.dt_uhma.setEnabled(flag_aero)
    gui.label_dt_uhma.setEnabled(flag_aero)
    gui.aero_start_label.setEnabled(flag_aero)
    gui.aero_start_offset.setEnabled(flag_aero)
    gui.aero_start_date.setEnabled(flag_aero)

    flag_mix_aero = flag.get("flag_mix_aero", 1) != 0
    gui.flag_mix_aero.setChecked(flag_mix_aero)

    aer_nucleation = aer.get("nucleation", True)
    gui.aer_nucleation.setChecked(aer_nucleation)

    aer_condensation = aer.get("condensation", True)
    gui.aer_condensation.setChecked(aer_condensation)

    aer_coagulation = aer.get("coagulation", True)
    gui.aer_coagulation.setChecked(aer_coagulation)

    aer_dry_deposition = aer.get("dry_deposition", False)
    gui.aer_dry_deposition.setChecked(aer_dry_deposition)

    # Note: Aerosol wet deposition is not yet implemented
    gui.aer_wet_deposition.setChecked(False)

    aer_snow_scavenge = aer.get("snow_scavenge", False)
    gui.aer_snow_scavenge.setChecked(aer_snow_scavenge)

    flag_vapor = flag.get("flag_vapor", 0) != 0
    gui.flag_vapor.setChecked(flag_vapor)

    flag_debug = flag.get("flag_debug", 0) != 0
    gui.flag_debug.setChecked(flag_debug)


def _update_gui_from_time_settings(settings, gui):
    time = settings.get("nml_time", dict())

    gui.start_date.setDateTime(
        QtCore.QDateTime.fromString(
            ",".join(
                f"{s:02}"
                for s in time.get("start_date", [2023, 4, 1, 0, 0, 0])
            ),
            "yyyy,MM,dd,HH,mm,ss",
        )
    )
    gui.end_date.setDateTime(
        QtCore.QDateTime.fromString(
            ",".join(
                f"{s:02}" for s in time.get("end_date", [2023, 4, 1, 0, 0, 0])
            ),
            "yyyy,MM,dd,HH,mm,ss",
        )
    )

    # Note: A negative value would be misinterpreted afterwards
    gui.trajectory_duration.setValue(
        abs(gui.start_date.dateTime().daysTo(gui.end_date.dateTime()))
    )
    gui.trajectory_duration.lastValue = gui.trajectory_duration.value()

    fullDays = gui.trajectory_duration.value()

    # Positive duration: end_date = ceil(start_date + full_days)
    gui.start_date.setEnabled(True)
    gui.end_date.setEnabled(False)

    gui.end_date.setDate(
        gui.start_date.date().addDays(
            fullDays + (gui.start_date.time() > QtCore.QTime(0, 0))
        )
    )

    aero_start_date = QtCore.QDateTime.fromString(
        ",".join(
            f"{s:02}"
            for s in time.get("aero_start_date", [2000, 1, 1, 0, 0, 0])
        ),
        "yyyy,MM,dd,HH,mm,ss",
    )
    secs_from_start = gui.start_date.dateTime().secsTo(aero_start_date)

    if secs_from_start < 0:
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setText("Eager early aerosols")
        msg.setInformativeText(
            "The aerosol simulation cannot start before the simulation itself."
            " Its starting time will be clamped to the simulation start."
        )
        msg.setWindowTitle("Warning loading INITFILE")
        msg.exec_()

    # Calculate aero_start_date based on aero_start_offset
    #  based on aero_start_date
    gui.aero_start_offset.setValue(secs_from_start // (60 * 60))
    gui.aero_start_date.setText(
        gui.start_date.dateTime()
        .addSecs(gui.aero_start_offset.value() * 60 * 60)
        .toString(" dd/MM/yyyy HH:mm:ss")
    )

    gui.dt_obs.setValue(float(time.get("dt_obs", 1800)))
    gui.dt_mete.setValue(float(time.get("dt_mete", 10)))
    gui.dt_emis.setValue(float(time.get("dt_emis", 60)))
    gui.dt_chem.setValue(float(time.get("dt_chem", 60)))
    gui.dt_depo.setValue(float(time.get("dt_depo", 60)))
    gui.dt_aero.setValue(float(time.get("dt_aero", 60)))
    gui.dt_uhma.setValue(float(time.get("dt_uhma", 10)))


def _update_gui_from_output_settings(settings, gui):
    output = settings.get("nml_output", dict())
    guis = settings.get("nml_gui", dict())

    gui.output_list_spc.setPlainText(
        ", ".join(
            s.strip()
            for s in output.get("output_list_spc", "").split(",")
            if len(s.strip()) > 0
        )
    )
    gui.output_list_emi.setPlainText(
        ", ".join(
            s.strip()
            for s in output.get("output_list_emi", "").split(",")
            if len(s.strip()) > 0
        )
    )
    gui.output_list_Vd.setPlainText(
        ", ".join(
            s.strip()
            for s in output.get("output_list_vd", "").split(",")
            if len(s.strip()) > 0
        )
    )
    gui.output_list_vap.setPlainText(
        ", ".join(
            s.strip()
            for s in output.get("output_list_vap", "").split(",")
            if len(s.strip()) > 0
        )
    )

    gui.description.setPlainText(guis.get("scenario_description", ""))


def _update_gui_from_compile_settings(settings, gui):
    guis = settings.get("nml_gui", dict())

    gui.sosaa_exe.setText(guis.get("sosaa_exe", "SOSAA.exe"))


def _update_gui_from_run_settings(settings, gui):
    guis = settings.get("nml_gui", dict())

    gui.launch_cmd.setText(
        guis.get("launch_cmd", "orterun --oversubscribe -n 4")
    )


def _update_gui_from_rsm_settings(settings, gui):
    rsm = settings.get("nml_rsm", dict())

    # Pretty-print the rsm path if relative to the working dir
    rsm_path = str(Path(rsm.get("rsm_path", "./sosaa-rsm.jl")).resolve())
    if rsm_path.startswith(str(Path.cwd())):
        rsm_path = f"./{Path(rsm_path).relative_to(Path.cwd())}"
    gui.rsm_path.setText(rsm_path)

    gui.rsm_train_seed.setText(rsm.get("train_seed", "my-train-seed"))
    gui.rsm_forest.setValue(rsm.get("forest_size", 16))
    gui.rsm_train_samples.setValue(rsm.get("train_samples", 1))

    # Pretty-print the rsm output path if relative to the working dir
    rsm_output = str(Path(rsm.get("rsm_output", "./rsm-output.jl")).resolve())
    if rsm_output.startswith(str(Path.cwd())):
        rsm_output = f"./{Path(rsm_output).relative_to(Path.cwd())}"
    gui.rsm_output.setText(rsm_output)

    gui.rsm_predict_seed.setText(rsm.get("predict_seed", "my-predict-seed"))
    gui.rsm_predict_samples.setValue(rsm.get("predict_samples", 1))

    gui.rsm_perturbation.setPlainText(
        rsm.get("predict_perturbation", "return inputs")
    )


def _update_gui_from_custom_settings(settings, gui):
    custom = settings.get("nml_custom", None)

    if custom is not None:
        i = 1

        for key, value in custom.items():
            getattr(gui, f"customKey_{i}").setText(str(key))
            getattr(gui, f"customVal_{i}").setText(str(value))

            i += 1

            if i > 30:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setText("Excessive custom config")
                msg.setInformativeText(
                    "The INITFILE contains more than 30 properties in the "
                    + f"'NML_CUSTOM' namelist. Not all of the {len(custom)} "
                    + "properties are shown in the GUI."
                )
                msg.setWindowTitle("Warning loading INITFILE")
                msg.exec_()

                break

        for i in range(i, 31):
            getattr(gui, f"customKey_{i}").setText("")
            getattr(gui, f"customVal_{i}").setText("")
    else:
        for i in range(1, 31):
            getattr(gui, f"customKey_{i}").setText("")
            getattr(gui, f"customVal_{i}").setText("")


def _update_gui_from_raw_settings(gui, raw):
    gui.rawEdit.document().setPlainText(raw)



"""
Clumped 0/1 sampler using a Markov Process

P(0) = p and P(1) = 1-p
clump = 0 => IID samples
clump -> 1 => highly correlated samples

"""


class Clump:
    def __init__(self, p=0.5, clump=0.0, rng=None):
        import numpy as np

        a = 1 - (1 - p) * (1 - clump)
        b = (1 - a) * p / (1 - p)

        self.C = np.array([[a, 1 - a], [b, 1 - b]])

        self.i = 0 if rng.random() < p else 1

    def sample(self, rng):
        p = self.C[self.i, 0]
        u = rng.random()

        self.i = 0 if u < p else 1

        return self.i

    def steady(self, X):
        import numpy as np

        return np.matmul(X, self.C)


def train_test_split(
    X, Y, test_size=0.25, random_state=None, shuffle=True, clump=0.0
):
    import numpy as np
    import pandas as pd

    assert len(X) == len(Y)
    assert type(X) == type(Y)
    assert test_size > 0.0
    assert test_size < 1.0
    assert random_state is not None
    assert clump >= 0.0
    assert clump < 1.0

    c = Clump(p=test_size, clump=clump, rng=random_state)

    if isinstance(X, pd.DataFrame):
        assert X.index.values.shape == Y.index.values.shape

        # Split only based on the first-level index instead of flattening
        n1 = len(X.index.levels[1])
        n0 = len(X) // n1

        C = np.array([c.sample(random_state) for _ in range(n0)])
        (I_train,) = np.nonzero(C)
        I_train = np.repeat(I_train, n1) * n1 + np.tile(
            np.arange(n1), len(I_train)
        )
        (I_test,) = np.nonzero(1 - C)
        I_test = np.repeat(I_test, n1) * n1 + np.tile(
            np.arange(n1), len(I_test)
        )
    else:
        C = np.array([c.sample(random_state) for _ in range(len(X))])
        (I_train,) = np.nonzero(C)
        (I_test,) = np.nonzero(1 - C)

    if shuffle:
        random_state.shuffle(I_train)
        random_state.shuffle(I_test)

    if isinstance(X, pd.DataFrame):
        X_train = X.iloc[I_train]
        X_test = X.iloc[I_test]

        Y_train = Y.iloc[I_train]
        Y_test = Y.iloc[I_test]
    else:
        X_train = X[I_train]
        X_test = X[I_test]

        Y_train = Y[I_train]
        Y_test = Y[I_test]

    return X_train, X_test, Y_train, Y_test



from . import IcarusPrediction


def analyse_icarus_predictions(
    predictions: IcarusPrediction,
    analysis,  #: Callable[[np.ndarray, np.ndarray,
    #                       np.random.Generator, dict,
    #                      ], np.ndarray],
    rng,  #: np.random.Generator,
    # number of samples to draw from expand each prediction per run
    n_uncertain_samples: int = 1,
    # number of repeats of the analysis to gather uncertainty
    n_analysis_runs: int = 100,
    **kwargs,
):
    import numpy as np

    progress = kwargs.get("progress", None)

    if progress is not None:
        progress.update_minor(
            value=0,
            min=0,
            max=n_analysis_runs,
            format="Monte Carlo Analysis Run %v/%m",
        )

    confidence = np.mean(predictions.confidence)

    results = []

    for _ in range(n_analysis_runs):
        confs = []
        preds = []
        for _ in range(n_uncertain_samples):
            I_conf = (
                rng.random(size=predictions.confidence.shape)
                <= predictions.confidence
            )
            (I_conf,) = np.nonzero(I_conf)

            confs.append(I_conf)
            preds.append(
                rng.normal(
                    loc=predictions.prediction[I_conf],
                    scale=predictions.uncertainty[I_conf],
                )
            )
        confs = np.concatenate(confs, axis=0)
        preds = np.concatenate(preds, axis=0)

        results.append(analysis(preds, confs, rng, **kwargs))

        if progress is not None:
            progress.update_minor()

    prediction = np.mean(np.stack(results, axis=0), axis=0)
    uncertainty = np.std(np.stack(results, axis=0), axis=0)

    return IcarusPrediction(
        prediction=prediction,
        uncertainty=uncertainty,
        confidence=confidence,
    )



import itertools


def generate_time_level_windows():
    # -0.5h, -1.5h, -3h, -6h, -12h, -24h, -48h
    # 0, -2, -5, -11, -23, -47, -95
    time_windows = [
        (0, 0),
        (-2, -1),
        (-5, -3),
        (-11, -6),
        (-23, -12),
        (-47, -24),
        (-95, -48),
    ]

    # +1l, +2l, +4l, +8l, +16l, +32l, +64
    top_windows = [(1, 1), (1, 2), (1, 4), (2, 8), (2, 16), (3, 32), (3, 64)]
    mid_windows = [(0, 0), (0, 0), (0, 0), (-1, 1), (-1, 1), (-2, 2), (-2, 2)]
    bot_windows = [
        (-1, -1),
        (-2, -1),
        (-4, -1),
        (-8, -2),
        (-16, -2),
        (-32, -3),
        (-64, -3),
    ]

    return list(
        itertools.chain(
            zip(time_windows, top_windows),
            zip(time_windows, mid_windows),
            zip(time_windows, bot_windows),
        )
    )


def generate_windowed_feature_names(columns):
    time_windows = ["-0.5h", "-1.5h", "-3h", "-6h", "-12h", "-24h", "-48h"]

    top_windows = ["+1l", "+2l", "+4l", "+8l", "+16l", "+32l", "+64l"]
    mid_windows = ["+0l", "+0l", "+0l", "±1l", "±1l", "±2l", "±2l"]
    bot_windows = ["-1l", "-2l", "-4l", "-8l", "-16l", "-32l", "-64l"]

    names = []

    for t, l in itertools.chain(
        zip(time_windows, top_windows),
        zip(time_windows, mid_windows),
        zip(time_windows, bot_windows),
    ):
        for c in columns:
            names.append(f"{c}{t}{l}")

    return names


def time_level_window_mean(input, t_range, l_range, progress=None):
    import numpy as np

    output = np.zeros(shape=input.shape)

    for t in range(input.shape[0]):
        mint = min(max(0, t + t_range[0]), input.shape[0])
        maxt = max(0, min(t + 1 + t_range[1], input.shape[0]))

        if mint == maxt:
            continue

        for lev in range(input.shape[1]):
            minl = min(max(0, lev + l_range[0]), input.shape[1])
            maxl = max(0, min(lev + 1 + l_range[1], input.shape[1]))

            if minl == maxl:
                continue

            output[t, lev, :] = np.mean(
                input[mint:maxt, minl:maxl, :], axis=(0, 1)
            )

    if progress is not None:
        progress.update_minor()

    return output



from pathlib import Path

from PyQt5 import QtWidgets


def browsePath(title, save=False, directory=False, origin=None):
    pathSelector = QtWidgets.QFileDialog()

    pathSelector.setAcceptMode(
        pathSelector.AcceptMode.AcceptSave
        if save
        else pathSelector.AcceptMode.AcceptOpen
    )

    pathSelector.setOption(pathSelector.DontUseNativeDialog)

    if not save:
        pathSelector.setFileMode(pathSelector.FileMode.ExistingFile)

    if directory:
        pathSelector.setFileMode(pathSelector.FileMode.Directory)
        pathSelector.setOption(pathSelector.ShowDirsOnly)
        pathSelector.setOption(pathSelector.DontResolveSymlinks)

    pathSelector.setWindowTitle(title)

    pathSelector.setDirectory(origin or str(Path.cwd()))

    if pathSelector.exec() == 1:
        path = Path(pathSelector.selectedFiles()[0]).resolve()
    else:
        path = None

    return path



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'about.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(683, 509)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.line = QtWidgets.QFrame(Dialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setWordWrap(True)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setText("")
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(Dialog)
        self.label_4.setWordWrap(True)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.label_5 = QtWidgets.QLabel(Dialog)
        self.label_5.setText("")
        self.label_5.setObjectName("label_5")
        self.verticalLayout.addWidget(self.label_5)
        self.label_6 = QtWidgets.QLabel(Dialog)
        self.label_6.setWordWrap(True)
        self.label_6.setObjectName("label_6")
        self.verticalLayout.addWidget(self.label_6)
        self.line_2 = QtWidgets.QFrame(Dialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.label_7 = QtWidgets.QLabel(Dialog)
        self.label_7.setWordWrap(True)
        self.label_7.setObjectName("label_7")
        self.verticalLayout.addWidget(self.label_7)
        self.frame = QtWidgets.QFrame(Dialog)
        self.frame.setMinimumSize(QtCore.QSize(0, 120))
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout.setProperty("margin", 0)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout.addItem(spacerItem)
        self.logo = QtWidgets.QLabel(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.logo.sizePolicy().hasHeightForWidth()
        )
        self.logo.setSizePolicy(sizePolicy)
        self.logo.setMinimumSize(QtCore.QSize(320, 120))
        self.logo.setMaximumSize(QtCore.QSize(320, 120))
        self.logo.setText("")
        self.logo.setObjectName("logo")
        self.horizontalLayout.addWidget(self.logo)
        spacerItem1 = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addWidget(self.frame)
        self.frame_2 = QtWidgets.QFrame(Dialog)
        self.frame_2.setMinimumSize(QtCore.QSize(0, 30))
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setProperty("margin", 0)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem2 = QtWidgets.QSpacerItem(
            160,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout_2.addItem(spacerItem2)
        self.okgreat = QtWidgets.QPushButton(self.frame_2)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.okgreat.sizePolicy().hasHeightForWidth()
        )
        self.okgreat.setSizePolicy(sizePolicy)
        self.okgreat.setMinimumSize(QtCore.QSize(0, 25))
        self.okgreat.setMaximumSize(QtCore.QSize(16777215, 25))
        self.okgreat.setObjectName("okgreat")
        self.horizontalLayout_2.addWidget(self.okgreat)
        spacerItem3 = QtWidgets.QSpacerItem(
            94,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout_2.addItem(spacerItem3)
        self.verticalLayout.addWidget(self.frame_2)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "About SOSAA"))
        self.label.setText(
            _translate(
                "Dialog",
                (
                    '<html><head/><body><p align="center"><span style="'
                    ' font-weight:700;">SOSAA</span></p><p'
                    ' align="center"><span style=" font-weight:700;">The model'
                    " to Simulate Organic vapours, Sulphuric Acid and"
                    " Aerosols</span></p></body></html>"
                ),
            )
        )
        self.label_2.setText(
            _translate(
                "Dialog",
                (
                    '<html><head/><body><p align="center">Multiscale Modelling'
                    " Group, Institute for Atmospheric and Earth System"
                    " Research (INAR), Helsinki, Finland</p></body></html>"
                ),
            )
        )
        self.label_4.setText(
            _translate(
                "Dialog",
                (
                    '<html><head/><body><p align="center">Development team'
                    ' (alphabetically): </p><p align="center">Michael Boy'
                    " (team leader), Dean Chen, Petri Clusius, Rosa Gierens,"
                    " Sampo Smolander, Ditte Taipale, Juniper Tyree, Luxi"
                    " Zhou, and Putian Zhou.</p></body></html>"
                ),
            )
        )
        self.label_6.setText(
            _translate(
                "Dialog",
                (
                    '<html><head/><body><p align="center">Contact the SOSAA'
                    ' team at <a href="mailto:sosaa@helsinki.fi"><span style="'
                    " text-decoration: underline;"
                    ' color:#007af4;">sosaa@helsinki.fi</span></a></p></body></html>'
                ),
            )
        )
        self.label_7.setText(
            _translate(
                "Dialog",
                (
                    '<html><head/><body><p align="center"><span style="'
                    ' font-weight:700;">LICENCE</span></p><p'
                    ' align="center">This program comes with ABSOLUTELY NO'
                    " WARRANTY. This is free software, and you are welcome to"
                    " redistribute it under certain conditions; see the"
                    " LICENCE file in the SOSAA GUI root"
                    " directory.</p></body></html>"
                ),
            )
        )
        self.okgreat.setText(_translate("Dialog", "OK"))


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



from __future__ import annotations

from ..icarus import IcarusPrediction, IcarusRSM


# SOSAA-RF with a percentile-based confidence score
class RandomForestSosaaRSM(IcarusRSM):
    def fit(
        self,
        X_train,  #: np.ndarray,
        Y_train,  #: np.ndarray,
        X_valid,  #: np.ndarray,
        Y_valid,  #: np.ndarray,
        rng,  #: np.random.Generator,
        n_trees: int = 16,
        progress=None,
        **kwargs,
    ) -> RandomForestSosaaRSM:
        import numpy as np
        from sklearn.covariance import EmpiricalCovariance
        from sklearn.decomposition import PCA
        from sklearn.ensemble import RandomForestRegressor

        # Fake use Y_valid
        Y_valid = Y_valid

        assert Y_train.shape[1:] == (1,)

        if progress is not None:
            progress.update_minor(
                value=0,
                min=0,
                max=5,
                format="Fitting the Truncated PCA OOD Detector",
            )

        self.pca = PCA(random_state=rng).fit(X_train)
        self.bn = np.searchsorted(
            np.cumsum(self.pca.explained_variance_ratio_), 0.95
        )

        if progress is not None:
            progress.update_minor(
                format="Fitting the Auto-Associative Error Covariance"
            )

        self.cov = EmpiricalCovariance().fit(
            (self._predict_truncated_pca(X_train) - X_train)
        )

        self.err_valid = np.sort(
            self.cov.mahalanobis(
                self._predict_truncated_pca(X_valid) - X_valid
            )
        )

        if progress is not None:
            progress.update_minor(
                format=(
                    "Training the Prediction Model and Uncertainty Quantifier"
                )
            )

        self.predictor = RandomForestRegressor(
            n_estimators=n_trees,
            random_state=rng,
            n_jobs=-1,
            min_samples_leaf=5,
            max_features=1.0 / 3.0,
        ).fit(X_train, Y_train.ravel())

        if progress is not None:
            progress.update_minor(format="Finished Training the SOSAA RSM")

        return self

    def predict(
        self,
        X_test,  #: np.ndarray,
        rng,  #: np.random.Generator,
        progress=None,
    ) -> IcarusPrediction:
        import joblib
        import numpy as np

        # No extra randomness is needed during prediction
        rng = rng

        if progress is not None:
            progress.update_minor(
                value=0,
                min=0,
                max=(1 + len(self.predictor.estimators_)),
                format="Generating the Confidence Scores",
            )

        confidence = 1.0 - np.searchsorted(
            self.err_valid,
            self.cov.mahalanobis(
                (self._predict_truncated_pca(X_test) - X_test)
            ),
        ) / len(self.err_valid)

        if progress is not None:
            progress.update_minor(
                format=(
                    "Generating"
                    f" {0}/{len(self.predictor.estimators_)} Ensemble"
                    " Predictions"
                )
            )

        def tree_predict(i, tree, X_test, progress=None) -> np.ndarray:
            result = tree.predict(X_test)

            if progress is not None:
                progress.update_minor(
                    format=(
                        "Generating"
                        f" {i}/{len(self.predictor.estimators_)} Ensemble"
                        " Predictions"
                    )
                )

            return result

        predictions = joblib.Parallel(n_jobs=-1, prefer="threads")(
            joblib.delayed(tree_predict)(i, tree, X_test, progress)
            for i, tree in enumerate(self.predictor.estimators_)
        )

        prediction = np.mean(np.stack(predictions, axis=0), axis=0).reshape(
            (len(X_test), 1)
        )
        uncertainty = np.std(np.stack(predictions, axis=0), axis=0).reshape(
            (len(X_test), 1)
        )

        return IcarusPrediction(
            prediction=prediction,
            uncertainty=uncertainty,
            confidence=confidence,
        )

    def _predict_truncated_pca(
        self,
        X,  #: np.ndarray
    ):  # -> np.ndarray:
        import numpy as np

        if self.pca.mean_ is not None:
            X = X - self.pca.mean_

        X_trans = np.dot(X, self.pca.components_[: self.bn].T)
        X = np.dot(X_trans, self.pca.components_[: self.bn])

        if self.pca.mean_ is not None:
            X = X + self.pca.mean_

        return X



from PyQt5 import QtCore

from ..settings import is_loading


def init_scenario_gui(gui):
    def changeTrajectoryDuration():
        if is_loading():
            return

        lastFullDays = gui.trajectory_duration.lastValue
        fullDays = gui.trajectory_duration.value()

        gui.trajectory_duration.lastValue = fullDays

        if fullDays < 0:
            # Negative duration: start_date = floor(end_date - full_days)
            gui.start_date.setEnabled(False)
            gui.end_date.setEnabled(True)

            if lastFullDays > 0:
                gui.end_date.setDate(gui.start_date.date())

            gui.start_date.setDate(gui.end_date.date().addDays(fullDays))
            gui.start_date.setTime(QtCore.QTime(0, 0))

            gui.aero_start_date.setText(
                gui.start_date.dateTime()
                .addSecs(gui.aero_start_offset.value() * 60 * 60)
                .toString(" dd/MM/yyyy HH:mm:ss")
            )
        elif fullDays > 0:
            # Positive duration: end_date = ceil(start_date + full_days)
            gui.start_date.setEnabled(True)
            gui.end_date.setEnabled(False)

            if lastFullDays < 0:
                gui.start_date.setDate(gui.end_date.date())

            gui.end_date.setDate(
                gui.start_date.date().addDays(
                    fullDays + (gui.start_date.time() > QtCore.QTime(0, 0))
                )
            )
            gui.end_date.setTime(QtCore.QTime(0, 0))
        else:
            # Zero duration, check which direction we are coming from

            if lastFullDays > 0:
                gui.end_date.setDate(
                    gui.start_date.date().addDays(
                        gui.start_date.time() > QtCore.QTime(0, 0)
                    )
                )
            else:
                gui.start_date.setDate(gui.end_date.date())

    gui.trajectory_duration.valueChanged.connect(changeTrajectoryDuration)
    gui.trajectory_duration.lastValue = gui.trajectory_duration.value()

    def changeAerosolStartOffset():
        if is_loading():
            return

        gui.aero_start_date.setText(
            gui.start_date.dateTime()
            .addSecs(gui.aero_start_offset.value() * 60 * 60)
            .toString(" dd/MM/yyyy HH:mm:ss")
        )

    gui.aero_start_offset.valueChanged.connect(changeAerosolStartOffset)

    def changeStartDate():
        if is_loading():
            return

        fullDays = gui.trajectory_duration.value()

        if fullDays >= 0:
            # Positive duration: end_date = ceil(start_date + full_days)
            gui.end_date.setDate(
                gui.start_date.date().addDays(
                    fullDays + (gui.start_date.time() > QtCore.QTime(0, 0))
                )
            )
            gui.end_date.setTime(QtCore.QTime(0, 0))

        gui.aero_start_date.setText(
            gui.start_date.dateTime()
            .addSecs(gui.aero_start_offset.value() * 60 * 60)
            .toString(" dd/MM/yyyy HH:mm:ss")
        )

    gui.start_date.dateTimeChanged.connect(changeStartDate)

    def changeEndDate():
        if is_loading():
            return

        fullDays = gui.trajectory_duration.value()

        if fullDays < 0:
            # Negative duration: start_date = floor(end_date - full_days)
            gui.start_date.setDate(gui.end_date.date().addDays(fullDays))
            gui.start_date.setTime(QtCore.QTime(0, 0))

            gui.aero_start_date.setText(
                gui.start_date.dateTime()
                .addSecs(gui.aero_start_offset.value() * 60 * 60)
                .toString(" dd/MM/yyyy HH:mm:ss")
            )

    gui.end_date.dateTimeChanged.connect(changeEndDate)



import os
import sys

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QCoreApplication

try:
    import platform

    # "Windows" / "Linux" / "Darwin"
    operating_system = platform.system() or "Linux"
except Exception:
    operating_system = "Linux"


# Set the GUI scaling factor for QT based on the operating system
def setup_qt_scaling():
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

    # Enable QT highdpi scaling and highdpi icons
    QtWidgets.QApplication.setAttribute(
        QtCore.Qt.AA_EnableHighDpiScaling, True
    )
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

    # Check if scaling is necessary, currently only required on Windows
    args = []

    for arg in sys.argv:
        args.append(arg.upper())

        if "--scaling_" in arg:
            os.environ["QT_SCALE_FACTOR"] = (
                f"{float(arg.replace('--scaling_', '')):3.2f}"
            )

            args.append("-NS")

    # FIXME petri: check if platform usually works for everyone
    if (
        os.name.upper() == "NT" or operating_system == "Windows"
    ) and "-NS" not in args:
        try:
            import ctypes

            sf = ctypes.windll.shcore.GetScaleFactorForDevice(0) / 100
            if sf < 1.3:
                sf = 1
            sf = 1 / sf

            u32 = ctypes.windll.user32
            scrhgt = u32.GetSystemMetrics(1)

            if scrhgt < 850:
                sf = sf * scrhgt / 850.0
        except Exception:
            sf = 1

            print(
                (
                    "Failed to get the scaling factor of the screen, falling"
                    f" back to {sf:3.2f}."
                ),
                flush=True,
            )

        os.environ["QT_SCALE_FACTOR"] = f"{sf:3.2f}"

    if "-NS" in args:
        print(
            (
                "Scaling factor overriden to"
                f" {os.environ['QT_SCALE_FACTOR']} from the commandline."
            ),
            flush=True,
        )


# Set the QT style for the GUI application
def setup_qt_style():
    styles = QtWidgets.QStyleFactory.keys()

    if "Fusion" in styles:
        QCoreApplication.instance().setStyle(
            QtWidgets.QStyleFactory.create("Fusion")
        )
    else:
        print(f"Available styles: {styles}", flush=True)



from PyQt5 import QtWidgets

from .ccn import init_plot as _init_ccn_plot
from .ccn import update_plot as _update_ccn_plot
from .diff import init_plot as _init_diff_plot
from .diff import update_plot as _update_diff_plot


def update_rsm_plots(gui):
    if gui.rsm_plots_dirty:
        _init_rsm_plots(gui)

        _update_ccn_plot(gui)
        _update_diff_plot(gui)

    gui.rsm_plots_dirty = False


def _init_rsm_plots(gui):
    if getattr(gui, "rsm_plot_init", False):
        return

    try:
        import matplotlib as mpl
        from matplotlib import pyplot as plt
        from matplotlib.backends.backend_qt5agg import (
            NavigationToolbar2QT as NavigationToolbar,
        )
    except ImportError as err:
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(f"Optional dependency {err.name} missing")
        msg.setInformativeText(
            "Please install sosaa-gui with the optional 'icarus' feature"
            " enabled."
        )
        msg.setWindowTitle("Missing optional dependency")
        msg.exec_()

        return

    mpl.use("Qt5Agg")
    plt.style.use("seaborn-v0_8")

    NavigationToolbar.toolitems.pop(-3)
    NavigationToolbar.toolitems.pop(-3)

    _init_ccn_plot(gui)
    _init_diff_plot(gui)

    gui.rsm_plot_init = True



from pathlib import Path


# Return the resolved absolute str(Path) for path relative to the package root
def resource_path(path):
    return str((Path(__file__).parent / path).resolve())


# open(file) for a file path relative to the package root
def open_resource(file, *args, **kwargs):
    return open(resource_path(file), *args, **kwargs)



import datetime

from ..netcdf import TrajectoryDatasets


def get_output_time(ds: TrajectoryDatasets):
    fdom = datetime.datetime.strptime(
        ds.out["time"].__dict__["first_day_of_month"],
        "%Y-%m-%d %H:%M:%S",
    )
    dt = (ds.date - fdom).total_seconds()

    out_t = ds.out["time"][:].data

    return out_t - dt



import os
from shutil import which

from PyQt5 import QtWidgets


def generateTerminalCommand(command: str, frameId):
    if which("urxvt") is not None:
        term = "urxvt"
        embed = "-embed"
    elif which("rxvt") is not None:
        term = "rxvt"
        embed = "-embed"
    elif which("uxterm") is not None:
        term = "uxterm"
        embed = "-into"
    elif which("xterm") is not None:
        term = "xterm"
        embed = "-into"
    else:
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText("Missing terminal")
        msg.setInformativeText(
            "Please install any of xterm, uxterm, rxvt, or urxvt."
        )
        msg.setWindowTitle("Error launching terminal")
        msg.exec_()

        return

    return term, [
        embed,
        f"{frameId}",
        "-sb",
        "-geometry",
        "640x480",
        "-hold",
        "-e",
        os.environ.get("SHELL", "sh"),
        "-x",
        "-c",
        command,
    ]



import hashlib


# https://stackoverflow.com/a/67809235
def df_to_numpy(df):
    try:
        shape = [len(level) for level in df.index.levels]
    except AttributeError:
        shape = [len(df.index)]
    ncol = df.shape[-1]
    if ncol > 1:
        shape.append(ncol)
    return df.to_numpy().reshape(shape)


def hash_for_dt(dt):
    if not (isinstance(dt, tuple) or isinstance(dt, list)):
        dt = [dt]

    dt_str = ".".join(dtt.strftime("%d.%m.%Y-%H:00%z") for dtt in dt)

    h = hashlib.shake_256()
    h.update(dt_str.encode("ascii"))

    return h



import os
from pathlib import Path

from PyQt5 import QtWidgets

from ..qt import operating_system


def init_gui_output(gui):
    def openOutputDirectory():
        output_dir = Path(gui.output_dir.text()).resolve()

        if not output_dir.exists():
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)

            if str(output_dir).startswith(str(Path.cwd())):
                output_dir = f"./{output_dir.relative_to(Path.cwd())}"

            msg.setText(f"The output directory '{output_dir}' does not exist.")
            msg.setWindowTitle("Invalid OUTPUT directory")
            msg.exec_()

            return

        if operating_system == "Windows":
            os.startfile(output_dir)
        if operating_system == "Linux":
            os.system(f'xdg-open "{output_dir}"')
        if operating_system == "Darwin":
            os.system('open "{output_dir}"')
        else:
            return

    gui.actionOpen_output_directory.triggered.connect(openOutputDirectory)

    def createOutputDirectories():
        output_dir = Path(gui.output_dir.text()).resolve()

        output_dir_str = str(output_dir)

        if output_dir_str.startswith(str(Path.cwd())):
            output_dir_str = f"./{output_dir.relative_to(Path.cwd())}"

        try:
            output_dir.mkdir(parents=True)

            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText(
                f"The output directory '{output_dir_str}' was created."
            )
            msg.setWindowTitle("Created OUTPUT directory")
            msg.exec_()
        except FileExistsError:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText(
                f"The output directory '{output_dir_str}' already existed."
            )
            msg.setWindowTitle("Created OUTPUT directory")
            msg.exec_()
        except Exception as e:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText(
                f"Failed to create the output directory '{output_dir_str}'."
            )
            msg.setInformativeText(str(e))
            msg.setWindowTitle("Invalid OUTPUT directory")
            msg.exec_()

    gui.actionCreate_output_directories.triggered.connect(
        createOutputDirectories
    )



from pathlib import Path

from PyQt5 import QtWidgets

from .worker import run_in_thread


def predict_sosaa_rsm(gui):
    # Disable concurrent training or prediction
    gui.rsm_build.setEnabled(False)
    gui.rsm_load.setEnabled(False)
    gui.rsm_predict.setEnabled(False)

    # Reset any previous RSM predictions
    gui.rsm_prediction = None

    # Generate the RSM predictions and update the GUI
    _generate_rsm_prediction_and_update_gui(gui)


def _generate_rsm_prediction_and_update_gui(gui):
    try:
        import numpy as np

        from ...sosaa_rsm.model.perturb import generate_perturbed_predictions

        # Configure the prediction
        n_samples = gui.rsm_predict_samples.value()
        prediction_path = Path(gui.rsm_output.text()).resolve()
        perturbation = _generate_perturbation_function(gui)

        predict_seed = np.random.SeedSequence(
            list(gui.rsm_predict_seed.text().encode())
        )
        predict_rng = np.random.RandomState(np.random.PCG64(predict_seed))

        if prediction_path.exists():
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Question)
            msg.setStandardButtons(
                QtWidgets.QMessageBox.Open
                | QtWidgets.QMessageBox.Reset
                | QtWidgets.QMessageBox.Cancel
            )
            msg.setText(
                "Do you want to load the existing RSM prediction or"
                " overwrite it?"
            )
            msg.setInformativeText(
                f"The file {str(prediction_path)} already exists."
            )
            msg.setWindowTitle("Existing SOSAA RSM Prediction")
            button = msg.exec_()

            if button == QtWidgets.QMessageBox.Cancel:
                return _on_prediction_finished(gui, err=None, result=None)

            overwrite_rsm_prediction = button != QtWidgets.QMessageBox.Open
        else:
            overwrite_rsm_prediction = True

        # Generate the RSM predictions in a worker thread
        gui.rsm_predict_thread = run_in_thread(
            lambda: generate_perturbed_predictions(
                gui.rsm_model,
                gui.rsm_dataset,
                predict_rng,
                n_samples,
                prediction_path,
                overwrite_rsm_prediction,
                perturbation,
                gui.rsm_predict_progress,
            ),
            lambda result: _on_prediction_finished(
                gui, err=None, result=result
            ),
            lambda err: _on_prediction_finished(gui, err),
        )
    except Exception as err:
        _on_prediction_finished(gui, err)


def _on_prediction_finished(gui, err, result=None):
    try:
        if err is not None:
            raise err

        # Generate the perturbed prediction using the RSM
        gui.rsm_prediction = result

        # Early return if loading or predicting was cancelled
        if gui.rsm_prediction is None:
            return
    except ImportError as err:
        # Gracefully catch missing dependencies
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(f"Optional dependency {err.name} missing")
        msg.setInformativeText(
            "Please install sosaa-gui with the optional 'icarus' feature"
            " enabled."
        )
        msg.setWindowTitle("Missing optional dependency")
        msg.exec_()

        return
    except Exception as err:
        # Gracefully catch internal errors
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(type(err).__name__)
        msg.setInformativeText(str(err))
        msg.setWindowTitle("Error predicting with SOSAA RSM")
        msg.exec_()

        return
    finally:
        gui.rsm_plots_dirty = True

        # Reset the RSM GUI to allow training a new RSM
        #  and making new predictions
        gui.rsm_predict_progress.update_major(value=0, format="")
        gui.rsm_predict_progress.update_minor(value=0, format="")

        gui.rsm_build.setEnabled(True)
        gui.rsm_load.setEnabled(True)
        gui.rsm_predict.setEnabled(True)

    # Communicate prediction success
    gui.rsm_predict_progress.update_major(
        value=1,
        max=1,
        format="Completed Predicting with the SOSAA RSM",
    )
    gui.rsm_predict_progress.update_minor(
        value=1,
        max=1,
        format=(
            f"The SOSAA RSM predictions are stored at {gui.rsm_output.text()}"
        ),
    )


def _generate_perturbation_function(gui):
    import pandas as pd

    perturbation_code = gui.rsm_perturbation.toPlainText()

    if len(perturbation_code) == 0:
        perturbation_code = "return inputs"

    # Source code for the perturbation 'sandbox'
    # The GUI specifies the body of the following function
    # def perturb_inputs(inputs: pandas.DataFrame) -> pandas.DataFrame:
    #     <BODY>
    perturbation_code = """import numpy
import pandas
import numpy as np
import pandas as pd

def perturb_inputs(inputs: pandas.DataFrame) -> pandas.DataFrame:
""" + "".join(f"    {line}" for line in perturbation_code.splitlines(True))

    def perturb_inputs_wrapper(inputs: pd.DataFrame) -> pd.DataFrame:
        locals = dict()
        globals = dict()

        exec(perturbation_code, globals, locals)

        return locals["perturb_inputs"](inputs)

    return perturb_inputs_wrapper



import re
import sys
import time
from io import StringIO

import f90nml
from f90nml.namelist import Namelist

from ..config import get_config, set_config
from ..resources import resource_path
from .from_file import update_gui_from_settings
from .from_gui import update_settings_from_gui

default_settings_path = resource_path("conf/defaults.init")
minimal_settings_path = resource_path("conf/minimal.init")


def is_loading():
    return globals().get("_is_loading", False)


def load_settings(gui, path):
    globals()["_is_loading"] = True

    print(f"Loading INITFILE from {path} ...", flush=True)

    with open(path, "r") as file:
        content = file.read()

    # Hack to ensure that _settings is not mistaken as a local variable
    globals()["_settings"] = f90nml.reads(
        _raw_setting_pattern.sub("", content)
    )

    # Configure the f90 namelist pretty-printing options
    _settings.indent = "  "
    _settings.end_comma = True
    _settings.uppercase = True

    # Extract and combine the raw inputs
    raw = "\n".join(m.group(1) for m in _raw_setting_pattern.finditer(content))

    update_gui_from_settings(_settings, gui, raw)

    display_settings(gui)

    globals()["_is_loading"] = False


def display_settings(gui):
    config = StringIO()

    _write_settings_to_file(gui, config, update_version=False)

    gui.currentInitFileContent.document().setPlainText(config.getvalue())


def print_settings(gui):
    print("\n" * 10, flush=True)

    _write_settings_to_file(gui, sys.stdout, update_version=True)
    sys.stdout.flush()

    display_settings(gui)


def save_settings(gui, path):
    print(f"Saving INITFILE to {path} ...", flush=True)

    with open(path, "w") as file:
        _write_settings_to_file(gui, file, update_version=True)

    display_settings(gui)


def _write_settings_to_file(gui, file, update_version=True):
    update_settings_from_gui(_settings, gui)

    # fmt: off
    file.write(f"! {'='*76} !\n")
    file.write(
        f"! {f'SOSAA INITFILE {_version_major}.{_version_minor}'.center(76)} !\n"  # noqa: E501
    )
    file.write(
        f"! {('Created at: ' + time.strftime('%B %d %Y, %H:%M:%S', time.localtime())).center(76)} !\n"  # noqa: E501
    )
    file.write(f"! {'='*76} !\n")
    # fmt: on

    if update_version:
        # Hack to ensure that _version_minor is not mistaken
        #  as a local variable
        globals()["_version_minor"] += 1

    file.write("\n")
    file.write(str(_settings))
    file.write("\n")

    file.write("\n")
    file.write(_raw_settings_header)
    file.write("\n")
    file.write(gui.rawEdit.toPlainText())
    file.write("\n")
    file.write(_raw_settings_footer)
    file.write("\n")


# Create the module-local settings global and its pretty-printing options
_settings = Namelist()
_settings.indent = "  "
_settings.end_comma = True
_settings.uppercase = True


# FIXME petri: only change the minor version when the configuration changes
_version_major = int(get_config("settings", "version", fallback="0")) + 1
set_config("settings", "version", str(_version_major))
_version_minor = 0


# Regex pattern to insert and extract the raw input from the GUI
_raw_settings_header = (
    f"! \\/ {'Raw input from the SOSAA GUI'.center(70, '-')} \\/ !"
)
_raw_settings_footer = (
    f"! /\\ {'Raw input from the SOSAA GUI'.center(70, '-')} /\\ !"
)
# fmt: off
_raw_setting_pattern = re.compile(
    rf"^{re.escape(_raw_settings_header)}$\n(.*?)\n^{re.escape(_raw_settings_footer)}$",  # noqa: E501
    flags=(re.MULTILINE | re.DOTALL),
)
# fmt: on



import configparser
import pathlib

from .resources import open_resource, resource_path

_CONFIG_PATH = "conf/config.ini"


# Get a configuration value from the INI file
def get_config(*args, **kwargs):
    gui_config = configparser.ConfigParser()

    if pathlib.Path(resource_path(_CONFIG_PATH)).exists():
        with open_resource(_CONFIG_PATH, "r") as file:
            gui_config.read_file(file)

    return gui_config.get(*args, **kwargs)


# Write a configuration value to the INI file
def set_config(section, option, value):
    gui_config = configparser.ConfigParser()

    if pathlib.Path(resource_path(_CONFIG_PATH)).exists():
        with open_resource(_CONFIG_PATH, "r") as file:
            gui_config.read_file(file)

    if not gui_config.has_section(section):
        gui_config.add_section(section)

    gui_config.set(section, option, value)

    with open_resource(_CONFIG_PATH, "w") as file:
        gui_config.write(file)


# Remove a configuration option from section in the INI file
def remove_config(section, option):
    if not pathlib.Path(resource_path(_CONFIG_PATH)).exists():
        return

    gui_config = configparser.ConfigParser()

    with open_resource(_CONFIG_PATH, "r") as file:
        gui_config.read_file(file)

    if gui_config.has_option(section, option):
        gui_config.remove_option(section, option)

    with open_resource(_CONFIG_PATH, "w") as file:
        gui_config.write(file)



from pathlib import Path

from PyQt5 import QtWidgets

from ..settings import (
    default_settings_path,
    display_settings,
    load_settings,
    minimal_settings_path,
    print_settings,
    save_settings,
)
from ..version import sosaa_version_pretty
from .browse import browsePath
from .style import buttonStyle
from .syntax import FortranNamelistHighlighter


def init_gui_loadsave(gui):
    _hideCurrentInitFile(gui)

    gui.actionOpen.triggered.connect(lambda: _loadSettings(gui))
    gui.loadButton.clicked.connect(lambda: _loadSettings(gui))

    def loadMinimalSettings():
        gui.currentInitFileToSave = None
        load_settings(gui, minimal_settings_path)
        _hideCurrentInitFile(gui)

    gui.actionLoad_minimal_settings.triggered.connect(loadMinimalSettings)

    # If no default settings exist, just use the minimal ones
    if Path(default_settings_path).exists():
        load_settings(gui, default_settings_path)
    else:
        load_settings(gui, minimal_settings_path)

    gui.actionPrint.triggered.connect(lambda: print_settings(gui))

    gui.saveButton.clicked.connect(lambda: _saveSettings(gui, path=None))
    gui.actionSave_2.triggered.connect(lambda: _saveSettings(gui, path=None))

    gui.saveCurrentButton.clicked.connect(
        lambda: _saveSettings(gui, path=gui.currentInitFileToSave)
    )
    gui.actionSave_to_current.triggered.connect(
        lambda: _saveSettings(gui, path=gui.currentInitFileToSave)
    )

    gui.saveDefaults.clicked.connect(
        lambda: _saveSettings(gui, path=default_settings_path)
    )

    # Set up drag and drop to load init files
    type(gui).dragEnterEvent = _dragInitFileEnter
    type(gui).dropEvent = _dropInitFile
    gui.setAcceptDrops(True)

    # Set up raw INITFILE syntax highlighting
    gui.rawEditHighlight = FortranNamelistHighlighter(
        gui, gui.rawEdit.document()
    )

    # Set up INITFILE update on switch to its tab
    gui.tabWidget.currentChanged.connect(lambda i: _tab_switched(gui, i))

    # Set up INITFILE syntax highlighting
    gui.currentInitFileContentHighlight = FortranNamelistHighlighter(
        gui,
        gui.currentInitFileContent.document(),
    )


def _loadSettings(gui):
    path = browsePath(title="Choose INITFILE")

    _loadSettingsWithPath(gui, path)


def _loadSettingsWithPath(gui, path):
    if (
        str(path) != default_settings_path
        and str(path) != minimal_settings_path
    ):
        gui.currentInitFileToSave = path
    else:
        gui.currentInitFileToSave = None

    if path is not None:
        load_settings(gui, path)

    if gui.currentInitFileToSave is None:
        _hideCurrentInitFile(gui)
    else:
        _showCurrentInitFile(gui, gui.currentInitFileToSave)


def _saveSettings(gui, path=None):
    if path is None:
        path = browsePath(title="Save INITFILE", save=True)

    if path is None:
        return

    if str(path) == minimal_settings_path:
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText("Invalid INITFILE path")
        msg.setInformativeText(
            "Overwriting the minimal INITFILE from the GUI is not allowed."
        )
        msg.setWindowTitle("Error saving INITFILE")
        msg.exec_()

        return

    save_settings(gui, path)

    if str(path) != default_settings_path:
        gui.currentInitFileToSave = path

        _showCurrentInitFile(gui, path)
    else:
        _hideCurrentInitFile(gui)


def _showCurrentInitFile(gui, path):
    gui.saveCurrentButton.setStyleSheet(buttonStyle("icons/save.svg"))
    gui.saveCurrentButton.setEnabled(True)
    gui.actionSave_to_current.setEnabled(True)

    pathstr = str(path)

    if str(path).startswith(str(Path.cwd())):
        path = path.relative_to(Path.cwd())
        pathstr = f"./{path}"

    gui.currentInitFile.setText(pathstr)

    gui.setWindowTitle(f"{sosaa_version_pretty}: {pathstr}")
    gui.saveCurrentButton.setToolTip(f"Save to {path.name}")


def _hideCurrentInitFile(gui):
    gui.saveCurrentButton.setEnabled(False)
    gui.actionSave_to_current.setEnabled(False)
    gui.saveCurrentButton.setStyleSheet(buttonStyle("icons/save-inactive.svg"))

    gui.currentInitFile.setText("")

    gui.setWindowTitle(sosaa_version_pretty)
    gui.saveCurrentButton.setToolTip("")


def _dragInitFileEnter(_gui, e):
    # Fake-access the gui variable
    _gui = _gui

    if e.mimeData().hasFormat("text/plain"):
        e.accept()
    else:
        e.ignore()


def _dropInitFile(gui, e):
    path = Path(e.mimeData().text().replace("file://", "").strip()).resolve()

    _loadSettingsWithPath(gui, path)


def _tab_switched(gui, _i):
    # Fake-access the i variable
    _i = _i

    if gui.tabWidget.currentWidget() == gui.initfile_tab:
        display_settings(gui)



from pathlib import Path

from PyQt5 import QtCore

from .terminal import generateTerminalCommand


def init_compile_gui(gui):
    gui.compile_start.setEnabled(True)
    gui.compile_clean.setEnabled(True)
    gui.compile_cleanchem.setEnabled(True)
    gui.compile_stop.setEnabled(False)

    def generateMakeVariables():
        return (
            [
                # fmt: off
                f"SOSAA_ROOT={Path(gui.main_dir.text()).resolve()}",
                f"CODE_DIR={Path(gui.main_dir.text()).resolve() / gui.code_dir.text()}",  # noqa: E501
                f"CHEMALL_DIR={Path(gui.main_dir.text()).resolve() / gui.chem_dir.text()}",  # noqa: E501
                f"CASE_DIR={Path(gui.main_dir.text()).resolve() / gui.case_dir.text()}",  # noqa: E501
                f"CHEM={gui.chemname_dir.text()}",
                f"CASE={gui.casename_dir.text()}",
                # fmt: on
            ]
            + (
                [f"ALT_NAME={gui.sosaa_exe.text()}"]
                if len(gui.sosaa_exe.text()) > 0
                else []
            )
            + (
                [f"INIT_FILE={Path(gui.currentInitFile.text()).resolve()}"]
                if len(gui.currentInitFile.text()) > 0
                else []
            )
        )

    def startCompilation():
        terminal = gui.terminal_compile

        gui.compile_start.setEnabled(False)
        gui.compile_clean.setEnabled(False)
        gui.compile_cleanchem.setEnabled(False)
        gui.compile_stop.setEnabled(True)

        if getattr(terminal, "process", None) is not None:
            return

        command = generateTerminalCommand(
            " ".join(
                [
                    "cd",
                    str(
                        Path(gui.main_dir.text()).resolve()
                        / gui.code_dir.text()
                    ),
                    "&&",
                    "make",
                    "sosaa.exe",
                ]
                + generateMakeVariables()
            ),
            int(terminal.winId()),
        )

        if command is None:
            gui.compile_start.setEnabled(True)
            gui.compile_clean.setEnabled(True)
            gui.compile_cleanchem.setEnabled(True)
            gui.compile_stop.setEnabled(False)

            return

        terminal.process = QtCore.QProcess(terminal)
        terminal.process.start(*command)

    gui.compile_start.clicked.connect(startCompilation)

    def stopCompilation():
        terminal = gui.terminal_compile

        gui.compile_start.setEnabled(True)
        gui.compile_clean.setEnabled(True)
        gui.compile_cleanchem.setEnabled(True)
        gui.compile_stop.setEnabled(False)

        if getattr(terminal, "process", None) is None:
            return

        terminal.process.kill()
        terminal.process = None

    gui.compile_stop.clicked.connect(stopCompilation)

    def cleanSosaa():
        terminal = gui.terminal_compile

        gui.compile_start.setEnabled(False)
        gui.compile_clean.setEnabled(False)
        gui.compile_cleanchem.setEnabled(False)
        gui.compile_stop.setEnabled(True)

        if getattr(terminal, "process", None) is not None:
            return

        command = generateTerminalCommand(
            " ".join(
                [
                    "cd",
                    str(
                        Path(gui.main_dir.text()).resolve()
                        / gui.code_dir.text()
                    ),
                    "&&",
                    "make",
                    "clean",
                ]
                + generateMakeVariables()
            ),
            int(terminal.winId()),
        )

        if command is None:
            gui.compile_start.setEnabled(True)
            gui.compile_clean.setEnabled(True)
            gui.compile_cleanchem.setEnabled(True)
            gui.compile_stop.setEnabled(False)

            return

        terminal.process = QtCore.QProcess(terminal)
        terminal.process.start(*command)

    gui.compile_clean.clicked.connect(cleanSosaa)

    def cleanChemistry():
        terminal = gui.terminal_compile

        gui.compile_start.setEnabled(False)
        gui.compile_clean.setEnabled(False)
        gui.compile_cleanchem.setEnabled(False)
        gui.compile_stop.setEnabled(True)

        if getattr(terminal, "process", None) is not None:
            return

        command = generateTerminalCommand(
            " ".join(
                [
                    "cd",
                    str(
                        Path(gui.main_dir.text()).resolve()
                        / gui.code_dir.text()
                    ),
                    "&&",
                    "make",
                    "cleanchem",
                ]
                + generateMakeVariables()
            ),
            int(terminal.winId()),
        )

        if command is None:
            gui.compile_start.setEnabled(True)
            gui.compile_clean.setEnabled(True)
            gui.compile_cleanchem.setEnabled(True)
            gui.compile_stop.setEnabled(False)

            return

        terminal.process = QtCore.QProcess(terminal)
        terminal.process.start(*command)

    gui.compile_cleanchem.clicked.connect(cleanChemistry)

    def recompile():
        gui.tabWidget.setCurrentWidget(gui.sosaa_tab)
        gui.sosaa_subtab.setCurrentWidget(gui.compile_tab)

        startCompilation()

    gui.recompile.clicked.connect(recompile)
    gui.actionRecompile_model.triggered.connect(recompile)



from collections import namedtuple
from functools import partial

from ..dataset import MLDataset
from ..icarus import IcarusPrediction, IcarusRSM
from ..icarus.analyse import analyse_icarus_predictions

TrainTestEvaluation = namedtuple(
    "TrainTestEvaluation",
    [
        "train_mse",
        "train_mae",
        "train_r2",
        "train_rmsce",
        "test_mse",
        "test_mae",
        "test_r2",
        "test_rmsce",
    ],
)


def analyse_train_test_perforance(
    model: IcarusRSM,
    dataset: MLDataset,
    rng,  #: np.random.Generator,
    n_samples: int,
    progress=None,
    **kwargs,
) -> TrainTestEvaluation:
    import numpy as np

    def mse_mae_analysis(Y_true, Y_pred, I_pred, rng, **kwargs):
        from sklearn.metrics import (
            mean_absolute_error,
            mean_squared_error,
            r2_score,
        )

        Y_true = dataset.Y_scaler.inverse_transform(Y_true[I_pred])
        Y_pred = dataset.Y_scaler.inverse_transform(Y_pred)

        mse = mean_squared_error(Y_true, Y_pred)
        mae = mean_absolute_error(Y_true, Y_pred)
        r2 = r2_score(Y_true, Y_pred)

        return np.array([mse, mae, r2])

    train_predictions = []
    for i in range(n_samples):
        if progress is not None:
            progress.update_major(
                format=f"Predicting on the Training Dataset {i}/{n_samples}"
            )

        train_predictions.append(
            model.predict(dataset.X_train, rng, progress=progress, **kwargs)
        )

    if progress is not None:
        progress.update_major(
            format="Combining the Predictions on the Training Dataset"
        )
        progress.update_minor(
            value=0,
            min=0,
            max=len(dataset.X_train),
            format="Training Prediction %v/%m",
        )

    combined_train_predictions = IcarusPrediction(
        prediction=[],
        uncertainty=[],
        confidence=[],
    )

    for i in range(len(dataset.X_train)):
        predictions = np.array([p.prediction[i] for p in train_predictions])
        uncertainties = np.array([p.uncertainty[i] for p in train_predictions])
        confidences = np.array([p.confidence[i] for p in train_predictions])

        def combine_predictions(Y_pred, I_pred, rng, **kwargs):
            return (
                np.mean(Y_pred)
                if len(Y_pred) > 0
                else np.mean(
                    train_predictions[
                        rng.choice(len(train_predictions))
                    ].prediction[i]
                )
            )

        cp = analyse_icarus_predictions(
            IcarusPrediction(
                prediction=predictions,
                uncertainty=uncertainties,
                confidence=confidences,
            ),
            combine_predictions,
            rng,
            n_uncertain_samples=1,
            n_analysis_runs=10,
            progress=None,
        )

        combined_train_predictions.prediction.append(cp.prediction)
        combined_train_predictions.uncertainty.append(cp.uncertainty)
        combined_train_predictions.confidence.append(cp.confidence)

        if progress is not None:
            progress.update_minor()

    train_predictions = IcarusPrediction(
        prediction=np.array(combined_train_predictions.prediction).reshape(
            -1, 1
        ),
        uncertainty=np.array(combined_train_predictions.uncertainty).reshape(
            -1, 1
        ),
        confidence=np.array(combined_train_predictions.confidence),
    )

    if progress is not None:
        progress.update_major(format="Evaluating on the Training Dataset")

    train_eval = analyse_icarus_predictions(
        train_predictions,
        partial(mse_mae_analysis, dataset.Y_train),
        rng,
        n_uncertain_samples=1,
        n_analysis_runs=10,
        progress=progress,
    )

    if progress is not None:
        progress.update_major(
            format="Calculating the Training Calibration Error"
        )

    train_rmsce = calculate_calibration_error(
        dataset.Y_scaler.inverse_transform(dataset.Y_train),
        dataset.Y_scaler.inverse_transform(train_predictions.prediction),
        train_predictions.uncertainty * dataset.Y_scaler.scale_,
        train_predictions.confidence,
        progress=progress,
    )

    test_predictions = []
    for i in range(n_samples):
        if progress is not None:
            progress.update_major(
                format=f"Predicting on the Test Dataset {i}/{n_samples}"
            )

        test_predictions.append(
            model.predict(dataset.X_test, rng, progress=progress, **kwargs)
        )

    if progress is not None:
        progress.update_major(
            format="Combining the Predictions on the Test Dataset"
        )

        progress.update_minor(
            value=0,
            min=0,
            max=len(dataset.X_test),
            format="Test Prediction %v/%m",
        )

    combined_test_predictions = IcarusPrediction(
        prediction=[],
        uncertainty=[],
        confidence=[],
    )

    for i in range(len(dataset.X_test)):
        predictions = np.array([p.prediction[i] for p in test_predictions])
        uncertainties = np.array([p.uncertainty[i] for p in test_predictions])
        confidences = np.array([p.confidence[i] for p in test_predictions])

        def combine_predictions(Y_pred, I_pred, rng, **kwargs):
            return (
                np.mean(Y_pred)
                if len(Y_pred) > 0
                else np.mean(
                    test_predictions[
                        rng.choice(len(test_predictions))
                    ].prediction[i]
                )
            )

        cp = analyse_icarus_predictions(
            IcarusPrediction(
                prediction=predictions,
                uncertainty=uncertainties,
                confidence=confidences,
            ),
            combine_predictions,
            rng,
            n_uncertain_samples=1,
            n_analysis_runs=10,
            progress=None,
        )

        combined_test_predictions.prediction.append(cp.prediction)
        combined_test_predictions.uncertainty.append(cp.uncertainty)
        combined_test_predictions.confidence.append(cp.confidence)

        if progress is not None:
            progress.update_minor()

    test_predictions = IcarusPrediction(
        prediction=np.array(combined_test_predictions.prediction).reshape(
            -1, 1
        ),
        uncertainty=np.array(combined_test_predictions.uncertainty).reshape(
            -1, 1
        ),
        confidence=np.array(combined_test_predictions.confidence),
    )

    if progress is not None:
        progress.update_major(format="Evaluating on the Test Dataset")

    test_eval = analyse_icarus_predictions(
        test_predictions,
        partial(mse_mae_analysis, dataset.Y_test),
        rng,
        n_uncertain_samples=1,
        n_analysis_runs=10,
        progress=progress,
    )

    if progress is not None:
        progress.update_major(format="Calculating the Test Calibration Error")

    test_rmsce = calculate_calibration_error(
        dataset.Y_scaler.inverse_transform(dataset.Y_test),
        dataset.Y_scaler.inverse_transform(test_predictions.prediction),
        test_predictions.uncertainty * dataset.Y_scaler.scale_,
        test_predictions.confidence,
        progress=progress,
    )

    return TrainTestEvaluation(
        train_mse=IcarusPrediction(
            prediction=train_eval.prediction[0],
            uncertainty=train_eval.uncertainty[0],
            confidence=train_eval.confidence,
        ),
        train_mae=IcarusPrediction(
            prediction=train_eval.prediction[1],
            uncertainty=train_eval.uncertainty[1],
            confidence=train_eval.confidence,
        ),
        train_r2=IcarusPrediction(
            prediction=train_eval.prediction[2],
            uncertainty=train_eval.uncertainty[2],
            confidence=train_eval.confidence,
        ),
        train_rmsce=train_rmsce,
        test_mse=IcarusPrediction(
            prediction=test_eval.prediction[0],
            uncertainty=test_eval.uncertainty[0],
            confidence=test_eval.confidence,
        ),
        test_mae=IcarusPrediction(
            prediction=test_eval.prediction[1],
            uncertainty=test_eval.uncertainty[1],
            confidence=test_eval.confidence,
        ),
        test_r2=IcarusPrediction(
            prediction=test_eval.prediction[2],
            uncertainty=test_eval.uncertainty[2],
            confidence=test_eval.confidence,
        ),
        test_rmsce=test_rmsce,
    )


def calculate_calibration_error(
    Y_true,  #: np.ndarray,
    Y_pred,  #: np.ndarray,
    Y_stdv,  #: np.ndarray,
    Y_conf,  #: np.ndarray,
    N: int = 1000,
    progress=None,
) -> IcarusPrediction:
    import numpy as np
    import scipy as sp

    sce = 0.0

    if progress is not None:
        progress.update_minor(
            value=0, min=0, max=N, format="Checking Percentile %p"
        )

    for i in range(N + 1):
        if progress is not None:
            progress.update_minor()

        p = i / N

        Yp = Y_pred.flatten() + Y_stdv.flatten() * sp.stats.norm.ppf(p)
        sce += (p - np.average(Y_true.flatten() < Yp, weights=Y_conf)) ** 2

    rmsce = np.sqrt(sce / N)

    return IcarusPrediction(
        prediction=rmsce,
        uncertainty=None,
        confidence=np.mean(Y_conf),
    )



import colorsys

from PyQt5.QtGui import QColor, QPalette


# Convert a HSL colour to an RNG hex string
def hsl_to_hex(hue, saturation, lightness):
    r, g, b = colorsys.hsv_to_rgb(hue / 360, saturation / 100, lightness / 100)

    return f"#{round(r*255):02X}{round(g*255):02X}{round(b*255):02X}"


# Generate a system of colours from a colour hue
def _coloursystem_from_hue(hue):
    return {
        "G0": hsl_to_hex(hue, 0, 0),
        "G10": hsl_to_hex(hue, 29, 14),
        "G20": hsl_to_hex(hue, 25, 21),
        "G30": hsl_to_hex(hue, 18, 26),
        "G40": hsl_to_hex(hue, 18, 33),
        "G50": hsl_to_hex(hue, 18, 40),
        "G60": hsl_to_hex(hue, 18, 46),
        "G70": hsl_to_hex(hue, 15, 54),
        "G80": hsl_to_hex(hue, 14, 66),
        "G90": hsl_to_hex(hue, 6, 69),
        "G100": hsl_to_hex(hue, 6, 74),
        "G110": hsl_to_hex(hue, 7, 80),
        "G120": hsl_to_hex(hue, 7, 82),
        "G130": hsl_to_hex(hue, 5, 88),
        "G140": hsl_to_hex(hue, 0, 98),
        "G150": hsl_to_hex(hue, 0, 100),
        "C0": hsl_to_hex(hue, 100, 0),
        "C10": hsl_to_hex(hue, 84, 15),
        "C20": hsl_to_hex(hue, 48, 28),
        "C30": hsl_to_hex(hue, 40, 36),
        "C40": hsl_to_hex(hue, 47, 39),
        "C50": hsl_to_hex(hue, 76, 42),
        "C60": hsl_to_hex(hue, 95, 41),
        "C70": hsl_to_hex(hue, 82, 53),
        "C80": hsl_to_hex(hue, 99, 61),
        "C90": hsl_to_hex(hue, 100, 73),
        "C100": hsl_to_hex(hue, 100, 81),
        "C110": hsl_to_hex(hue, 85, 87),
        "C120": hsl_to_hex(hue, 100, 90),
        "C130": hsl_to_hex(hue, 100, 93),
        "C140": hsl_to_hex(hue, 100, 98),
        "C150": hsl_to_hex(hue, 100, 100),
    }


# Generate a light/dark QPalette from a colour hue
def get_style_palette(hue, dark):
    coloursystem = _coloursystem_from_hue(hue)

    palette = QPalette()

    if dark:
        palette.setColor(QPalette.Window, QColor(coloursystem["G30"]))
        palette.setColor(QPalette.WindowText, QColor(coloursystem["G150"]))
        palette.setColor(QPalette.Base, QColor(coloursystem["G20"]))
        palette.setColor(QPalette.AlternateBase, QColor(coloursystem["G30"]))
        palette.setColor(QPalette.ToolTipBase, QColor(coloursystem["G10"]))
        palette.setColor(QPalette.ToolTipText, QColor(coloursystem["G150"]))
        palette.setColor(QPalette.Text, QColor(coloursystem["G150"]))
        palette.setColor(QPalette.Button, QColor(coloursystem["G30"]))
        palette.setColor(QPalette.ButtonText, QColor(coloursystem["G150"]))
        palette.setColor(QPalette.BrightText, QColor(coloursystem["C150"]))
        palette.setColor(QPalette.Link, QColor(coloursystem["C70"]))
        palette.setColor(QPalette.Highlight, QColor(coloursystem["G120"]))
        palette.setColor(QPalette.HighlightedText, QColor(coloursystem["G20"]))
        palette.setColor(
            QPalette.Active, QPalette.Button, QColor(coloursystem["G50"])
        )
        palette.setColor(
            QPalette.Disabled,
            QPalette.ButtonText,
            QColor(coloursystem["G100"]),
        )
        palette.setColor(
            QPalette.Disabled,
            QPalette.WindowText,
            QColor(coloursystem["G100"]),
        )
        palette.setColor(
            QPalette.Disabled, QPalette.Text, QColor(coloursystem["G100"])
        )
        palette.setColor(
            QPalette.Disabled, QPalette.Light, QColor(coloursystem["G30"])
        )
    else:
        palette.setColor(QPalette.Window, QColor(coloursystem["G120"]))
        palette.setColor(QPalette.WindowText, QColor(coloursystem["G0"]))
        palette.setColor(QPalette.Base, QColor(coloursystem["G130"]))
        palette.setColor(QPalette.AlternateBase, QColor(coloursystem["G120"]))
        palette.setColor(QPalette.ToolTipBase, QColor(coloursystem["G140"]))
        palette.setColor(QPalette.ToolTipText, QColor(coloursystem["G0"]))
        palette.setColor(QPalette.Text, QColor(coloursystem["G0"]))
        palette.setColor(QPalette.Button, QColor(coloursystem["G120"]))
        palette.setColor(QPalette.ButtonText, QColor(coloursystem["G0"]))
        palette.setColor(QPalette.BrightText, QColor(coloursystem["C10"]))
        palette.setColor(QPalette.Link, QColor(coloursystem["C80"]))
        palette.setColor(QPalette.Highlight, QColor(coloursystem["C30"]))
        palette.setColor(
            QPalette.HighlightedText, QColor(coloursystem["G130"])
        )
        palette.setColor(
            QPalette.Active, QPalette.Button, QColor(coloursystem["G100"])
        )
        palette.setColor(
            QPalette.Disabled, QPalette.ButtonText, QColor(coloursystem["G50"])
        )
        palette.setColor(
            QPalette.Disabled, QPalette.WindowText, QColor(coloursystem["G50"])
        )
        palette.setColor(
            QPalette.Disabled, QPalette.Text, QColor(coloursystem["G50"])
        )
        palette.setColor(
            QPalette.Disabled, QPalette.Light, QColor(coloursystem["G120"])
        )

    return palette



from PyQt5 import QtGui, QtWidgets

from ..layouts import gui
from ..resources import resource_path
from ..version import sosaa_version_pretty
from .compile import init_compile_gui
from .dirs import init_dirs_gui
from .help import init_gui_help
from .loadsave import init_gui_loadsave
from .modules import init_modules_gui
from .output import init_gui_output
from .rsm import init_rsm_gui
from .run import init_run_gui
from .scenario import init_scenario_gui
from .style import init_gui_style


class QtSosaaGui(gui.Ui_MainWindow, QtWidgets.QMainWindow):
    """Main program window."""

    def __init__(self):
        super(QtSosaaGui, self).__init__()
        self.setupUi(self)

        self.currentInitFileToSave = None

        self.setWindowTitle(sosaa_version_pretty)
        self.setWindowIcon(QtGui.QIcon(resource_path("icons/sosaa-icon.svg")))

        self.actionQuit_Ctrl_Q.triggered.connect(self.close)

        init_gui_style(self)
        init_gui_loadsave(self)
        init_gui_output(self)
        init_gui_help(self)

        init_dirs_gui(self)
        init_modules_gui(self)
        init_scenario_gui(self)
        init_compile_gui(self)
        init_run_gui(self)
        init_rsm_gui(self)



import darkdetect
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QCoreApplication

from ..config import get_config, remove_config, set_config
from ..resources import resource_path
from ..style import get_style_palette, hsl_to_hex
from .syntax import STYLES


def init_gui_style(gui):
    # Style can be EITHER system, dark, or light
    gui.themeGroup = QtWidgets.QActionGroup(gui)
    gui.themeGroup.addAction(gui.actionSystem)
    gui.themeGroup.addAction(gui.actionDark)
    gui.themeGroup.addAction(gui.actionLight)

    # Open a colour picker window to select the new style colour
    def actionStyleColourTrigger():
        colourSelector = QtWidgets.QColorDialog(gui)
        colourSelector.setOption(
            QtWidgets.QColorDialog.ColorDialogOption.DontUseNativeDialog
        )
        colourSelector.setOption(
            QtWidgets.QColorDialog.ColorDialogOption.NoButtons
        )

        def colourChanged(colour):
            gui.hue = colour.getHsl()[0]
            set_config("style", "hue", str(gui.hue))
            _setLightDarkStyle(gui, gui.dark)

        colourSelector.setCurrentColor(
            QtGui.QColor(hsl_to_hex(gui.hue, 100, 100))
        )
        colourSelector.currentColorChanged.connect(colourChanged)

        colourSelector.exec_()

    gui.actionChange_Colour.triggered.connect(actionStyleColourTrigger)

    _loadDefaultStyle(gui)

    def actionSystemTrigger(checked):
        if checked:
            set_config("style", "theme", "system")
            _setLightDarkStyle(gui, darkdetect.isDark())

    gui.actionSystem.triggered.connect(actionSystemTrigger)

    def actionLightTrigger(checked):
        if checked:
            set_config("style", "theme", "light")
            _setLightDarkStyle(gui, False)

    gui.actionLight.triggered.connect(actionLightTrigger)

    def actionDarkTrigger(checked):
        if checked:
            set_config("style", "theme", "dark")
            _setLightDarkStyle(gui, True)

    gui.actionDark.triggered.connect(actionDarkTrigger)

    def actionResetStyleTrigger():
        remove_config("style", "hue")
        remove_config("style", "theme")

        _loadDefaultStyle(gui)

    gui.actionReset_Style.triggered.connect(actionResetStyleTrigger)

    _loadDefaultFont(gui)

    def actionSetGlobalFontTrigger():
        fontSelector = QtWidgets.QFontDialog()
        font, ok = fontSelector.getFont(gui.font(), parent=gui)

        if ok:
            set_config("style", "font", font.toString())
            _refresh_font(gui, font)

    gui.actionSet_Global_Font.triggered.connect(actionSetGlobalFontTrigger)

    def actionResetFontTrigger():
        remove_config("style", "font")

        _loadDefaultFont(gui)

    gui.actionReset_Fonts.triggered.connect(actionResetFontTrigger)


def _loadDefaultStyle(gui):
    gui.hue = int(get_config("style", "hue", fallback="316"))

    theme = get_config("style", "theme", fallback="system")

    if theme == "light":
        gui.actionLight.setChecked(True)
        _setLightDarkStyle(gui, False)
    elif theme == "dark":
        gui.actionDark.setChecked(True)
        _setLightDarkStyle(gui, True)
    else:
        gui.actionSystem.setChecked(True)
        _setLightDarkStyle(gui, darkdetect.isDark())


def _loadDefaultFont(gui):
    font_str = get_config("style", "font", fallback=None)

    if font_str is None:
        font = QtGui.QFont()
        font.setBold(False)
        font.setItalic(False)
        font.setWeight(50)
    else:
        font = QtGui.QFont()
        font.fromString(font_str)

    _refresh_font(gui, font)


def _setLightDarkStyle(gui, dark):
    gui.dark = dark

    QCoreApplication.instance().setPalette(
        get_style_palette(gui.hue, gui.dark)
    )

    _refresh_style(gui)


def _refresh_style(gui):
    gui.saveCurrentButton.setStyleSheet(
        buttonStyle(
            "icons/save-inactive.svg"
            if gui.currentInitFileToSave is None
            else "icons/save.svg"
        )
    )
    gui.saveButton.setStyleSheet(buttonStyle("icons/save-as.svg"))
    gui.loadButton.setStyleSheet(buttonStyle("icons/load.svg"))
    gui.saveDefaults.setStyleSheet(buttonStyle("icons/save-defaults.svg"))
    gui.recompile.setStyleSheet(buttonStyle("icons/recompile.svg"))

    gui.rsm_perturbation_header.setText(
        QtCore.QCoreApplication.translate(
            "MainWindow",
            (
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"'
                ' "http://www.w3.org/TR/REC-html40/strict.dtd">\n'
                '<html><head><meta name="qrichtext" content="1" /><meta'
                ' charset="utf-8" /><style type="text/css">\n'
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:'Segoe UI';"
                ' font-size:9pt; font-weight:400; font-style:normal;">\n'
                '<p style=" margin-top:0px; margin-bottom:0px;'
                " margin-left:0px; margin-right:0px; -qt-block-indent:0;"
                ' text-indent:0px;"><span style=" font-weight:700;'
                f' color:\'{STYLES["keyword"][2+gui.dark]}\';">def</span>'
                ' <span style=" font-weight:700;'
                f' color:\'{STYLES["defclass"][2+gui.dark]}\';">'
                'perturb_inputs</span><span style=" font-weight:700;'
                f' color:\'{STYLES["brace"][2+gui.dark]}\';">(</span>inputs:'
                ' pandas<span style=" font-weight:700;'
                f' color:\'{STYLES["operator"][2+gui.dark]}\';">.</span>'
                'DataFrame<span style=" font-weight:700;'
                f' color:\'{STYLES["brace"][2+gui.dark]}\';">)</span> <span'
                ' style=" font-weight:700;'
                f' color:\'{STYLES["operator"][2+gui.dark]}\';">-&gt;</span>'
                ' pandas<span style=" font-weight:700;'
                f' color:\'{STYLES["operator"][2+gui.dark]}\';">.</span>'
                "DataFrame:</p></body></html>"
            ),
        )
    )

    if getattr(gui, "rawEditHighlight", None) is not None:
        gui.rawEditHighlight.rehighlight()
        gui.currentInitFileContentHighlight.rehighlight()
        gui.rsm_perturbation_highlight.rehighlight()


def buttonStyle(icon):
    icon_path_escaped = resource_path(icon).replace("\\", "\\\\")

    return (
        f"background-image: url('{icon_path_escaped}');\nbackground-repeat:"
        " no-repeat;"
    )


def _refresh_font(gui, font):
    gui.setFont(font)

    gui.menuFile.setFont(font)
    gui.menuTools.setFont(font)
    gui.menuSettings.setFont(font)
    gui.menuStyle.setFont(font)
    gui.menuFont.setFont(font)
    gui.menuHelp.setFont(font)

    monospace = QtGui.QFont()
    monospace.fromString(font.toString())
    monospace.setFamily("Courier New")
    monospace.setStyleHint(QtGui.QFont.StyleHint.Monospace)

    gui.rawEdit.setFont(monospace)
    gui.currentInitFileContent.setFont(monospace)
    gui.rsm_perturbation.setFont(monospace)



import datetime
from collections import namedtuple
from pathlib import Path

from ..dataset import load_and_cache_dataset
from ..icarus import IcarusRSM

TrainTestEvaluation = namedtuple(
    "TrainTestEvaluation",
    [
        "train_mse",
        "train_mae",
        "train_r2",
        "train_rmsce",
        "test_mse",
        "test_mae",
        "test_r2",
        "test_rmsce",
    ],
)


def train_and_cache_model(
    dt: datetime.datetime,
    clump: float,
    datasets: dict,
    models: dict,
    cls,
    rng,  #: np.random.RandomState,
    input_dir: Path,
    output_dir: Path,
    model_path: Path,
    overwrite_model: bool,
    progress=None,
    **kwargs,
) -> IcarusRSM:
    import joblib

    if isinstance(dt, tuple) or isinstance(dt, list):
        dt = tuple(sorted(dt))

    model_key = (cls.__name__, dt, clump)

    cached = models.get(model_key)

    if cached is not None:
        return cached

    if Path(model_path).exists() and not overwrite_model:
        model = joblib.load(model_path)

        models[model_key] = model

        return model

    dataset = load_and_cache_dataset(
        dt, clump, datasets, input_dir, output_dir, progress=progress
    )

    model = cls().fit(
        X_train=dataset.X_train,
        Y_train=dataset.Y_train,
        X_valid=dataset.X_valid,
        Y_valid=dataset.Y_valid,
        rng=rng,
        progress=progress,
        **kwargs,
    )

    joblib.dump(model, model_path)

    models[model_key] = model

    return model



import signal
import time

from PyQt5 import QtWidgets

from sosaa_gui.gui import QtSosaaGui
from sosaa_gui.qt import setup_qt_scaling, setup_qt_style
from sosaa_gui.version import sosaa_version_pretty

# Enable GUI termination using ctrl-c
signal.signal(signal.SIGINT, signal.SIG_DFL)

print(
    (
        f"{sosaa_version_pretty} started at:"
        f" {time.strftime('%B %d %Y, %H:%M:%S', time.localtime())}"
    ),
    flush=True,
)

setup_qt_scaling()

app = QtWidgets.QApplication([])

gui = QtSosaaGui()
gui.setGeometry(30, 30, 900, 700)
gui.show()

setup_qt_style()

app.exec_()



import datetime
from collections import namedtuple
from pathlib import Path

TrajectoryPaths = namedtuple(
    "TrajectoryPaths",
    [
        "date",
        "out",
        "aer",
        "ant",
        "bio",
        "met",
    ],
)


def get_sosaa_dataset_paths(
    dt: datetime.datetime, input_dir: Path, output_dir: Path
) -> TrajectoryPaths:
    out_path = output_dir / "output.nc"
    # fmt: on
    aer_path = (
        input_dir
        / f"{dt.strftime('%Y%m%d')}_7daybwd_Hyde_traj_AER_{24-dt.hour:02}_L3.nc"  # noqa: E501
    )
    ant_path = (
        input_dir
        / f"{dt.strftime('%Y%m%d')}_7daybwd_Hyde_traj_ANT_{24-dt.hour:02}_L3.nc"  # noqa: E501
    )
    bio_path = (
        input_dir
        / f"{dt.strftime('%Y%m%d')}_7daybwd_Hyde_traj_BIO_{24-dt.hour:02}_L3.nc"  # noqa: E501
    )
    # fmt: off
    met_path = input_dir / f"METEO_{dt.strftime('%Y%m%d')}_R{24-dt.hour:02}.nc"

    out_path = out_path.resolve()
    aer_path = aer_path.resolve()
    ant_path = ant_path.resolve()
    bio_path = bio_path.resolve()
    met_path = met_path.resolve()

    for file in [out_path, aer_path, ant_path, bio_path, met_path]:
        if not file.exists():
            raise FileNotFoundError(str(file))

    return TrajectoryPaths(
        date=dt,
        out=out_path,
        aer=aer_path,
        ant=ant_path,
        bio=bio_path,
        met=met_path,
    )



from ...labels.time import get_output_time
from ...netcdf import TrajectoryDatasets


def interpolate_aerosol_emissions(ds: TrajectoryDatasets, key: str):
    import scipy as sp

    out_t = get_output_time(ds)
    out_h = ds.out["lev"][:].data

    aer_t = ds.aer["time"][:].data
    aer_h = ds.aer["mid_layer_height"][:].data

    aer_t_h = ds.aer[key][:].T

    aer_t_h_int = sp.interpolate.interp2d(
        x=aer_h,
        y=aer_t,
        z=aer_t_h,
        kind="linear",
        bounds_error=False,
        fill_value=0.0,
    )

    return aer_t_h_int(x=out_h, y=out_t)


def get_aer_emissions_features(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    return pd.DataFrame(
        {
            "time": np.repeat(get_output_time(ds), ds.out["lev"].shape[0]),
            "level": np.tile(ds.out["lev"][:].data, ds.out["time"].shape[0]),
            "aer_3_10_nm": interpolate_aerosol_emissions(
                ds, "3-10nm"
            ).flatten(),
            "aer_10_20_nm": interpolate_aerosol_emissions(
                ds, "10-20nm"
            ).flatten(),
            "aer_20_30_nm": interpolate_aerosol_emissions(
                ds, "20-30nm"
            ).flatten(),
            "aer_30_50_nm": interpolate_aerosol_emissions(
                ds, "30-50nm"
            ).flatten(),
            "aer_50_70_nm": interpolate_aerosol_emissions(
                ds, "50-70nm"
            ).flatten(),
            "aer_70_100_nm": interpolate_aerosol_emissions(
                ds, "70-100nm"
            ).flatten(),
            "aer_100_200_nm": interpolate_aerosol_emissions(
                ds, "100-200nm"
            ).flatten(),
            "aer_200_400_nm": interpolate_aerosol_emissions(
                ds, "200-400nm"
            ).flatten(),
            "aer_400_1000_nm": interpolate_aerosol_emissions(
                ds, "400-1000nm"
            ).flatten(),
        }
    ).set_index(["time", "level"])



import datetime
from pathlib import Path

from PyQt5 import QtWidgets

from .worker import run_in_thread


def build_sosaa_rsm(gui, rsm_should_exist: bool):
    # Disable concurrent training or prediction
    gui.rsm_build.setEnabled(False)
    gui.rsm_load.setEnabled(False)
    gui.rsm_predict.setEnabled(False)

    # Reset the RSM evaluation metrics
    gui.rsm_train_mse.setText("")
    gui.rsm_train_mae.setText("")
    gui.rsm_train_r2.setText("")
    gui.rsm_train_rmsce.setText("")
    gui.rsm_test_mse.setText("")
    gui.rsm_test_mae.setText("")
    gui.rsm_test_r2.setText("")
    gui.rsm_test_rmsce.setText("")

    # Reset the RSM training results
    gui.rsm_dataset = None
    gui.rsm_model = None
    gui.rsm_prediction = None

    # Build the RSM, evaluate it, and update the GUI
    _train_evaluate_sosaa_rsm(gui, rsm_should_exist)


def _train_evaluate_sosaa_rsm(gui, rsm_should_exist: bool):
    try:
        import numpy as np

        # SOSAA-RF has slightly more accurate predictions,
        #  has better performance, and works well with the
        #  percentile-based OOD scorer.
        # Future work can take the thesis evaluation
        #  experiment as inspiration for how to specialise
        #  support for SOSAA-PADRE-RF (e.g. with direct
        #  perturbation predictions)
        # from ...sosaa_rsm.rsms.sosaa_padre_rf import (
        #     PairwiseDifferenceRegressionRandomForestSosaaRSM as SosaaRSM,
        # )
        from ...sosaa_rsm.rsms.sosaa_rf import RandomForestSosaaRSM as SosaaRSM

        # Configure the RSM
        input_dir = Path(gui.input_dir.text()).resolve()
        output_dir = Path(gui.output_dir.text()).resolve()
        rsm_path = Path(gui.rsm_path.text()).resolve()
        dt = gui.end_date.dateTime()
        dt = datetime.datetime(
            year=dt.date().year(),
            month=dt.date().month(),
            day=dt.date().day(),
            hour=dt.time().hour(),
            minute=dt.time().minute(),
            second=dt.time().second(),
        )
        gui.rsm_dt = dt
        clump = 0.75  # sensible default
        n_trees = gui.rsm_forest.value()
        n_samples = gui.rsm_train_samples.value()

        train_seed = np.random.SeedSequence(
            list(gui.rsm_train_seed.text().encode())
        )
        train_rng = np.random.RandomState(np.random.PCG64(train_seed))
        eval_rng = np.random.RandomState(np.random.PCG64(train_seed))

        overwrite_rsm = False

        if rsm_path.exists():
            if not rsm_should_exist:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Question)
                msg.setStandardButtons(
                    QtWidgets.QMessageBox.Open
                    | QtWidgets.QMessageBox.Reset
                    | QtWidgets.QMessageBox.Cancel
                )
                msg.setText(
                    "Do you want to load the existing RSM or overwrite it?"
                )
                msg.setInformativeText(
                    f"The file {str(rsm_path)} already exists."
                )
                msg.setWindowTitle("Existing SOSAA RSM")
                button = msg.exec_()

                if button == QtWidgets.QMessageBox.Cancel:
                    return _on_build_finished(
                        gui, rsm_should_exist, err=None, result=None
                    )

                overwrite_rsm = button != QtWidgets.QMessageBox.Open
        elif rsm_should_exist:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Question)
            msg.setStandardButtons(
                QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.Cancel
            )
            msg.setText("Do you want to train a new SOSAA RSM?")
            msg.setInformativeText(f"The file {str(rsm_path)} does not exist.")
            msg.setWindowTitle("Missing SOSAA RSM")
            button = msg.exec_()

            if button == QtWidgets.QMessageBox.Cancel:
                return _on_build_finished(
                    gui, rsm_should_exist, err=None, result=None
                )

        # Build and evaluate the RSM in a worker thread
        gui.rsm_build_thread = run_in_thread(
            lambda: _train_evaluate_sosaa_rsm_job(
                gui,
                overwrite_rsm,
                input_dir,
                output_dir,
                rsm_path,
                dt,
                clump,
                SosaaRSM,
                n_trees,
                n_samples,
                train_rng,
                eval_rng,
            ),
            lambda result: _on_build_finished(
                gui, rsm_should_exist, err=None, result=result
            ),
            lambda err: _on_build_finished(gui, rsm_should_exist, err),
        )
    except Exception as err:
        _on_build_finished(gui, rsm_should_exist, err)


def _train_evaluate_sosaa_rsm_job(
    gui,
    overwrite_rsm,
    input_dir,
    output_dir,
    rsm_path,
    dt,
    clump,
    cls,
    n_trees,
    n_samples,
    train_rng,
    eval_rng,
):
    from ...sosaa_rsm.dataset import load_and_cache_dataset
    from ...sosaa_rsm.model import train_and_cache_model
    from ...sosaa_rsm.model.evaluate import analyse_train_test_perforance

    gui.rsm_build_progress.update_major(
        value=0, min=0, max=6 + 2 * n_samples, format="Training the SOSAA RSM"
    )

    # Datasets and models are not cached between runs
    datasets = dict()
    models = dict()

    # Train the RSM model
    model = train_and_cache_model(
        dt,
        clump,
        datasets,
        models,
        cls,
        train_rng,
        input_dir,
        output_dir,
        rsm_path,
        overwrite_rsm,
        n_trees=n_trees,
        n_samples=n_samples,
        progress=gui.rsm_build_progress,
    )

    # Early return if loading or training was cancelled
    if model is None:
        return

    gui.rsm_build_progress.update_major(format="Loading the SOSAA Dataset")

    # Load the dataset again
    #  only needed if the model was loaded from disk
    dataset = load_and_cache_dataset(
        dt,
        clump,
        datasets,
        input_dir,
        output_dir,
        progress=gui.rsm_build_progress,
    )

    # Evaluate the model's train/test performance
    train_test_eval = analyse_train_test_perforance(
        model,
        dataset,
        eval_rng,
        n_samples,
        progress=gui.rsm_build_progress,
    )

    return (model, dataset, train_test_eval)


def _on_build_finished(gui, rsm_should_exist: bool, err, result=None):
    try:
        from ...sosaa_rsm.icarus import IcarusPrediction

        if err is not None:
            raise err

        # Early return if loading or training was cancelled
        if result is None:
            return

        model, dataset, train_test_eval = result
        gui.rsm_dataset = dataset
        gui.rsm_model = model

        # Pretty-print an Icarus prediction
        def fip(p: IcarusPrediction):
            if p.uncertainty is not None:
                return (
                    f"({p.prediction:.02} ± {p.uncertainty:.02}) |"
                    f" {p.confidence:.02}"
                )
            else:
                return f"{p.prediction:.02} | {p.confidence:.02}"

        # Output the new model evaluation results
        gui.rsm_train_mse.setText(fip(train_test_eval.train_mse))
        gui.rsm_train_mae.setText(fip(train_test_eval.train_mae))
        gui.rsm_train_r2.setText(fip(train_test_eval.train_r2))
        gui.rsm_train_rmsce.setText(fip(train_test_eval.train_rmsce))
        gui.rsm_test_mse.setText(fip(train_test_eval.test_mse))
        gui.rsm_test_mae.setText(fip(train_test_eval.test_mae))
        gui.rsm_test_r2.setText(fip(train_test_eval.test_r2))
        gui.rsm_test_rmsce.setText(fip(train_test_eval.test_rmsce))
    except ImportError as err:
        # Gracefully catch missing dependencies
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(f"Optional dependency {err.name} missing")
        msg.setInformativeText(
            "Please install sosaa-gui with the optional 'icarus' feature"
            " enabled."
        )
        msg.setWindowTitle("Missing optional dependency")
        msg.exec_()

        return
    except Exception as err:
        # Gracefully catch internal errors
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(type(err).__name__)
        msg.setInformativeText(str(err))
        msg.setWindowTitle("Error training SOSAA RSM")
        msg.exec_()

        return
    finally:
        gui.rsm_plots_dirty = True

        # Reset the RSM GUI to allow training a new RSM
        gui.rsm_build_progress.update_major(
            value=0, format="No SOSAA RSM is loaded"
        )
        gui.rsm_build_progress.update_minor(value=0, format="")

        gui.rsm_build.setEnabled(True)
        gui.rsm_load.setEnabled(True)

    # Communicate training success and allow RSM predictions
    gui.rsm_predict.setEnabled(True)

    gui.rsm_build_progress.update_major(
        value=1,
        max=1,
        format="Completed {} the SOSAA RSM".format(
            "loading" if rsm_should_exist else "training"
        ),
    )
    gui.rsm_build_progress.update_minor(
        value=1,
        max=1,
        format=f"The SOSAA RSM is stored at {gui.rsm_path.text()}",
    )



from pathlib import Path

from PyQt5 import QtCore, QtWidgets

from .terminal import generateTerminalCommand


def init_run_gui(gui):
    gui.model_start.setEnabled(True)
    gui.model_stop.setEnabled(False)

    def startModelRun():
        terminal = gui.terminal_run

        gui.model_start.setEnabled(False)
        gui.model_stop.setEnabled(True)

        if getattr(terminal, "process", None) is not None:
            return

        command = generateTerminalCommand(
            " ".join(
                [
                    "cd",
                    str(
                        Path(gui.main_dir.text()).resolve()
                        / gui.case_dir.text()
                        / gui.casename_dir.text()
                    ),
                    "&&",
                    gui.launch_cmd.text(),
                    f"./{gui.sosaa_exe.text()}",
                    str(Path(gui.currentInitFile.text()).resolve()),
                ]
            ),
            int(terminal.winId()),
        )

        if command is None:
            gui.compile_start.setEnabled(True)
            gui.compile_clean.setEnabled(True)
            gui.compile_cleanchem.setEnabled(True)
            gui.compile_stop.setEnabled(False)

            return

        if not gui.saveCurrentButton.isEnabled():
            gui.model_start.setEnabled(True)
            gui.model_stop.setEnabled(False)

            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Unsaved INITFILE")
            msg.setInformativeText(
                "The INITFILE must be saved before starting a new SOSAA run."
            )
            msg.setWindowTitle("Error starting SOSAA")
            msg.exec_()

            return

        gui.actionSave_to_current.trigger()

        terminal.process = QtCore.QProcess(terminal)
        terminal.process.start(*command)

    gui.model_start.clicked.connect(startModelRun)

    def forceStopModelRun():
        terminal = gui.terminal_run

        gui.model_start.setEnabled(True)
        gui.model_stop.setEnabled(False)

        if getattr(terminal, "process", None) is None:
            return

        terminal.process.kill()
        terminal.process = None

    gui.model_stop.clicked.connect(forceStopModelRun)



from pathlib import Path

from PyQt5 import QtWidgets

from .browse import browsePath


def init_dirs_gui(gui):
    def changeMainDirectory():
        path = browsePath(title="Choose the main directory", directory=True)

        if path is None:
            return

        if str(path).startswith(str(Path.cwd())):
            path = f"./{path.relative_to(Path.cwd())}"

        gui.main_dir.setText(str(path))

    gui.browse_main.clicked.connect(changeMainDirectory)

    def changeCodeDirectory():
        main_dir = Path(gui.main_dir.text()).resolve()

        path = browsePath(
            title="Choose the source code directory",
            directory=True,
            origin=str(main_dir),
        )

        if path is None:
            return

        if not str(path).startswith(str(main_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid source code directory")
            msg.setInformativeText(
                "The source code directory must be inside the main directory."
            )
            msg.setWindowTitle("Error selecting directory")
            msg.exec_()

            return

        path = path.relative_to(main_dir)

        gui.code_dir.setText(str(path))

    gui.browse_code.clicked.connect(changeCodeDirectory)

    def changeChemistryDirectory():
        main_dir = Path(gui.main_dir.text()).resolve()

        path = browsePath(
            title="Choose the chemistry directory",
            directory=True,
            origin=str(main_dir),
        )

        if path is None:
            return

        if not str(path).startswith(str(main_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid chemistry directory")
            msg.setInformativeText(
                "The chemistry directory must be inside the main directory."
            )
            msg.setWindowTitle("Error selecting directory")
            msg.exec_()

            return

        path = path.relative_to(main_dir)

        gui.chem_dir.setText(str(path))

    gui.browse_chem.clicked.connect(changeChemistryDirectory)

    def changeChemistryNameDirectory():
        chem_dir = (
            Path(gui.main_dir.text()).resolve() / gui.chemistry_dir.text()
        )

        path = browsePath(
            title="Choose the chemistry name directory",
            directory=True,
            origin=str(chem_dir),
        )

        if path is None:
            return

        if not str(path).startswith(str(chem_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid chemistry name directory")
            msg.setInformativeText(
                "The chemistry name directory must be inside the chemistry"
                " directory."
            )
            msg.setWindowTitle("Error selecting directory")
            msg.exec_()

            return

        path = path.relative_to(chem_dir)

        gui.chemname_dir.setText(str(path))

    gui.browse_chemname.clicked.connect(changeChemistryNameDirectory)

    def changeCaseDirectory():
        main_dir = Path(gui.main_dir.text()).resolve()

        path = browsePath(
            title="Choose the case directory",
            directory=True,
            origin=str(main_dir),
        )

        if path is None:
            return

        if not str(path).startswith(str(main_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid case directory")
            msg.setInformativeText(
                "The case directory must be inside the main directory."
            )
            msg.setWindowTitle("Error selecting directory")
            msg.exec_()

            return

        path = path.relative_to(main_dir)

        gui.case_dir.setText(str(path))

    gui.browse_case.clicked.connect(changeCaseDirectory)

    def changeCaseNameDirectory():
        case_dir = Path(gui.main_dir.text()).resolve() / gui.case_dir.text()

        path = browsePath(
            title="Choose the case name directory",
            directory=True,
            origin=str(case_dir),
        )

        if path is None:
            return

        if not str(path).startswith(str(case_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid case name directory")
            msg.setInformativeText(
                "The case name directory must be inside the case directory."
            )
            msg.setWindowTitle("Error selecting directory")
            msg.exec_()

            return

        path = path.relative_to(case_dir)

        gui.casename_dir.setText(str(path))

    gui.browse_casename.clicked.connect(changeCaseNameDirectory)

    def changeSosaaExeFile():
        case_dir = (
            Path(gui.main_dir.text()).resolve()
            / gui.case_dir.text()
            / gui.casename_dir.text()
        )

        path = browsePath(
            title="Choose the SOSAA executable file", origin=str(case_dir)
        )

        if path is None:
            return

        if not str(path).startswith(str(case_dir)):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Invalid SOSAA executable file")
            msg.setInformativeText(
                "The SOSAA executable file must be inside the case name"
                " directory."
            )
            msg.setWindowTitle("Error selecting SOSAA executable")
            msg.exec_()

            return

        path = path.relative_to(case_dir)

        gui.sosaa_exe.setText(str(path))

    gui.browse_exe.clicked.connect(changeSosaaExeFile)

    def changeOutputDirectory():
        path = browsePath(title="Choose the output directory", directory=True)

        if path is None:
            return

        if str(path).startswith(str(Path.cwd())):
            path = f"./{path.relative_to(Path.cwd())}"

        gui.output_dir.setText(str(path))

    gui.browse_output.clicked.connect(changeOutputDirectory)

    def changeOutputPath():
        gui.currentOutputPath.setText(gui.output_dir.text())

    gui.output_dir.textChanged.connect(changeOutputPath)

    def changeInputDirectory():
        path = browsePath(title="Choose the input directory", directory=True)

        if path is None:
            return

        if str(path).startswith(str(Path.cwd())):
            path = f"./{path.relative_to(Path.cwd())}"

        gui.input_dir.setText(str(path))

    gui.browse_input.clicked.connect(changeInputDirectory)



from ..netcdf import TrajectoryDatasets
from .time import get_output_time


def get_ccn_concentration(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    (ccn_bin_indices,) = np.nonzero(ds.out["dp_dry_fs"][:].data > 80e-9)
    ccn_concentration = np.sum(
        ds.out["nconc_par"][:].data[:, ccn_bin_indices, :], axis=1
    )

    return pd.DataFrame(
        {
            "time": np.repeat(get_output_time(ds), ds.out["lev"].shape[0]),
            "level": np.tile(ds.out["lev"][:].data, ds.out["time"].shape[0]),
            "ccn": ccn_concentration.flatten(),
        }
    ).set_index(["time", "level"])



from __future__ import annotations

import abc
from collections import namedtuple

IcarusPrediction = namedtuple(
    "IcarusPrediction",
    [
        "prediction",
        "uncertainty",
        "confidence",
    ],
)


class IcarusRSM(abc.ABC):
    @abc.abstractmethod
    def fit(
        self,
        X_train,  #: np.ndarray,
        Y_train,  #: np.ndarray,
        X_valid,  #: np.ndarray,
        Y_valid,  #: np.ndarray,
        rng,  #: np.random.Generator,
        **kwargs,
    ) -> IcarusRSM:
        return self

    @abc.abstractmethod
    def predict(
        self,
        X_test,  #: np.ndarray,
        rng,  #: np.random.Generator,
        **kwargs,
    ) -> IcarusPrediction:
        return None



import datetime
from collections import namedtuple
from pathlib import Path

from .clump import train_test_split
from .features import get_features_from_raw_features
from .features.extract import get_raw_features_for_dataset
from .labels import get_labels_for_dataset
from .netcdf import load_trajectory_dataset
from .paths import get_sosaa_dataset_paths
from .utils import hash_for_dt

MLDataset = namedtuple(
    "MLDataset",
    [
        "date",
        "paths",
        "X_raw",
        "Y_raw",
        "X_train",
        "X_valid",
        "X_test",
        "Y_train",
        "Y_valid",
        "Y_test",
        "X_scaler",
        "Y_scaler",
    ],
)


def load_and_cache_dataset(
    dt: datetime.datetime,
    clump: float,
    datasets: dict,
    input_dir: Path,
    output_dir: Path,
    progress=None,
) -> MLDataset:
    import numpy as np
    import pandas as pd
    from sklearn.preprocessing import StandardScaler

    if isinstance(dt, tuple) or isinstance(dt, list):
        dt = tuple(sorted(dt))

    cached = datasets.get((dt, clump))

    if cached is not None:
        return cached

    if isinstance(dt, tuple) or isinstance(dt, list):
        mls = [
            load_and_cache_dataset(
                dtt, clump, datasets, input_dir, output_dir, progress=progress
            )
            for dtt in dt
        ]

        dp = tuple(ml.paths for ml in mls)
        X_raw = pd.concat([ml.X_raw for ml in mls], axis="index")
        Y = pd.concat([ml.Y_raw for ml in mls], axis="index")

        train_features = np.concatenate(
            [ml.X_scaler.inverse_transform(ml.X_train) for ml in mls], axis=0
        )
        train_labels = np.concatenate(
            [ml.Y_scaler.inverse_transform(ml.Y_train) for ml in mls], axis=0
        )
        valid_features = np.concatenate(
            [ml.X_scaler.inverse_transform(ml.X_valid) for ml in mls], axis=0
        )
        valid_labels = np.concatenate(
            [ml.Y_scaler.inverse_transform(ml.Y_valid) for ml in mls], axis=0
        )
        test_features = np.concatenate(
            [ml.X_scaler.inverse_transform(ml.X_test) for ml in mls], axis=0
        )
        test_labels = np.concatenate(
            [ml.Y_scaler.inverse_transform(ml.Y_test) for ml in mls], axis=0
        )
    else:
        dp = get_sosaa_dataset_paths(dt, input_dir, output_dir)
        ds = load_trajectory_dataset(dp)

        X_raw = get_raw_features_for_dataset(ds)

        X = get_features_from_raw_features(X_raw, progress=progress)
        Y = np.log10(get_labels_for_dataset(ds) + 1)

        rng = np.random.RandomState(
            seed=int.from_bytes(hash_for_dt(dt).digest(4), "little")
        )

        (
            train_features,
            test_features,
            train_labels,
            test_labels,
        ) = train_test_split(
            X,
            Y,
            test_size=0.25,
            random_state=rng,
            clump=clump,
        )
        (
            train_features,
            valid_features,
            train_labels,
            valid_labels,
        ) = train_test_split(
            train_features,
            train_labels,
            test_size=1.0 / 3.0,
            random_state=rng,
            clump=clump,
        )

        # Close the NetCDF datasets
        ds.out.close()
        ds.aer.close()
        ds.ant.close()
        ds.bio.close()
        ds.met.close()

    # Scale features to N(0,1)
    # - only fit on training data
    # - OOD inputs for constants at training time are blown up
    feature_scaler = StandardScaler().fit(train_features)
    feature_scaler.scale_[np.nonzero(feature_scaler.var_ == 0.0)] = (
        np.nan_to_num(np.inf)
    )

    label_scaler = StandardScaler().fit(train_labels)

    train_features = feature_scaler.transform(train_features)
    train_labels = label_scaler.transform(train_labels)
    valid_features = feature_scaler.transform(valid_features)
    valid_labels = label_scaler.transform(valid_labels)
    test_features = feature_scaler.transform(test_features)
    test_labels = label_scaler.transform(test_labels)

    dataset = MLDataset(
        date=dt,
        paths=dp,
        X_raw=X_raw,
        Y_raw=Y,
        X_train=train_features,
        X_valid=valid_features,
        X_test=test_features,
        Y_train=train_labels,
        Y_valid=valid_labels,
        Y_test=test_labels,
        X_scaler=feature_scaler,
        Y_scaler=label_scaler,
    )

    datasets[(dt, clump)] = dataset

    return dataset



from ...labels.time import get_output_time
from ...netcdf import TrajectoryDatasets


def interpolate_meteorology_values(ds: TrajectoryDatasets, key: str):
    import scipy as sp

    out_t = get_output_time(ds)
    out_h = ds.out["lev"][:].data

    met_t = ds.met["time"][:].data
    met_h = ds.met["lev"][:].data

    met_t_h = ds.met[key][:]

    met_t_h_int = sp.interpolate.interp2d(
        x=met_h,
        y=met_t,
        z=met_t_h,
        kind="linear",
        bounds_error=False,
        fill_value=0.0,
    )

    return met_t_h_int(x=out_h, y=out_t)


def interpolate_meteorology_time_values(ds: TrajectoryDatasets, key: str):
    import numpy as np
    import scipy as sp

    out_t = get_output_time(ds)
    out_h = ds.out["lev"][:].data

    met_t = ds.met["time"][:].data

    met_t_v = ds.met[key][:]

    met_t_int = sp.interpolate.interp1d(
        x=met_t,
        y=met_t_v,
        kind="linear",
        bounds_error=False,
        fill_value=0.0,
    )

    return np.repeat(
        met_t_int(x=out_t).reshape(-1, 1),
        out_h.shape[0],
        axis=1,
    )


def get_meteorology_features(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    return pd.DataFrame(
        {
            "time": np.repeat(get_output_time(ds), ds.out["lev"].shape[0]),
            "level": np.tile(ds.out["lev"][:].data, ds.out["time"].shape[0]),
            "met_t": interpolate_meteorology_values(ds, "t").flatten(),
            "met_q": interpolate_meteorology_values(ds, "q").flatten(),
            "met_ssr": interpolate_meteorology_time_values(
                ds, "ssr"
            ).flatten(),
            "met_lsm": interpolate_meteorology_time_values(
                ds, "lsm"
            ).flatten(),
            "met_blh": interpolate_meteorology_time_values(
                ds, "blh"
            ).flatten(),
        }
    ).set_index(["time", "level"])



from ..expand import (
    generate_time_level_windows,
    generate_windowed_feature_names,
    time_level_window_mean,
)
from ..utils import df_to_numpy


def get_features_from_raw_features(raw_features, progress=None):
    import joblib
    import numpy as np
    import pandas as pd

    raw_features_np = df_to_numpy(raw_features)

    if progress is not None:
        progress.update_minor(
            value=0,
            min=0,
            max=len(generate_time_level_windows()),
            format="Expanding space-time %v/%m",
        )

    features_np = np.concatenate(
        [
            raw_features.index.get_level_values(0)
            .to_numpy()
            .reshape(
                (
                    raw_features.index.levels[0].size,
                    raw_features.index.levels[1].size,
                    1,
                )
            ),
            raw_features.index.get_level_values(1)
            .to_numpy()
            .reshape(
                (
                    raw_features.index.levels[0].size,
                    raw_features.index.levels[1].size,
                    1,
                )
            ),
        ]
        + joblib.Parallel(n_jobs=-1, prefer="threads")(
            [
                joblib.delayed(time_level_window_mean)(
                    raw_features_np, t, l, progress=progress
                )
                for t, l in generate_time_level_windows()
            ]
        ),
        axis=2,
    )

    # Trim off the first two days, for which the time features are ill-defined
    features_np_trimmed = features_np[95:-1, :, :]

    feature_names = ["time", "level"] + generate_windowed_feature_names(
        raw_features.columns
    )

    features = pd.DataFrame(
        features_np_trimmed.reshape(
            features_np_trimmed.shape[0] * features_np_trimmed.shape[1],
            features_np_trimmed.shape[2],
        ),
        columns=feature_names,
    ).set_index(["time", "level"])

    return features



from ...labels.time import get_output_time
from ...netcdf import TrajectoryDatasets


def interpolate_biogenic_emissions(ds: TrajectoryDatasets, key: str):
    import numpy as np
    import scipy as sp

    out_t = get_output_time(ds)
    out_h = ds.out["lev"][:].data

    # depth of each box layer, assuming level heights are midpoints
    #  and end points are clamped
    out_d = (
        np.array(list(out_h[1:]) + [out_h[-1]])
        - np.array([out_h[0]] + list(out_h[:-1]))
    ) / 2.0

    bio_t = ds.bio["time"][:].data

    # Biogenic emissions are limited to boxes at <= 10m height
    biogenic_emission_layers = np.nonzero(out_h <= 10.0)
    biogenic_emission_layer_height_cumsum = np.cumsum(
        out_d[biogenic_emission_layers]
    )
    biogenic_emission_layer_proportion = (
        biogenic_emission_layer_height_cumsum
        / biogenic_emission_layer_height_cumsum[-1]
    )
    num_biogenic_emission_layers = sum(out_h <= 10.0)

    bio_t_h = np.zeros(shape=(out_t.size, out_h.size))

    bio_t_int = sp.interpolate.interp1d(
        x=bio_t,
        y=ds.bio[key][:],
        kind="linear",
        bounds_error=False,
        fill_value=0.0,
    )

    # Split up the biogenic emissions relative to the depth of the boxes
    bio_t_h[:, biogenic_emission_layers] = (
        np.tile(bio_t_int(x=out_t), (num_biogenic_emission_layers, 1, 1))
        * biogenic_emission_layer_proportion.reshape(-1, 1, 1)
    ).T

    return bio_t_h


def get_bio_emissions_features(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    return pd.DataFrame(
        {
            "time": np.repeat(get_output_time(ds), ds.out["lev"].shape[0]),
            "level": np.tile(ds.out["lev"][:].data, ds.out["time"].shape[0]),
            "bio_acetaldehyde": interpolate_biogenic_emissions(
                ds, "acetaldehyde"
            ).flatten(),
            "bio_acetone": interpolate_biogenic_emissions(
                ds, "acetone"
            ).flatten(),
            "bio_butanes_and_higher_alkanes": interpolate_biogenic_emissions(
                ds, "butanes-and-higher-alkanes"
            ).flatten(),
            "bio_butanes_and_higher_alkenes": interpolate_biogenic_emissions(
                ds, "butenes-and-higher-alkenes"
            ).flatten(),
            "bio_ch4": interpolate_biogenic_emissions(ds, "CH4").flatten(),
            "bio_co": interpolate_biogenic_emissions(ds, "CO").flatten(),
            "bio_ethane": interpolate_biogenic_emissions(
                ds, "ethane"
            ).flatten(),
            "bio_ethanol": interpolate_biogenic_emissions(
                ds, "ethanol"
            ).flatten(),
            "bio_ethene": interpolate_biogenic_emissions(
                ds, "ethene"
            ).flatten(),
            "bio_formaldehyde": interpolate_biogenic_emissions(
                ds, "formaldehyde"
            ).flatten(),
            "bio_hydrogen_cyanide": interpolate_biogenic_emissions(
                ds, "hydrogen-cyanide"
            ).flatten(),
            "bio_iosprene": interpolate_biogenic_emissions(
                ds, "isoprene"
            ).flatten(),
            "bio_mbo": interpolate_biogenic_emissions(ds, "MBO").flatten(),
            "bio_methanol": interpolate_biogenic_emissions(
                ds, "methanol"
            ).flatten(),
            "bio_methyl_bromide": interpolate_biogenic_emissions(
                ds, "methyl-bromide"
            ).flatten(),
            "bio_methyl_chloride": interpolate_biogenic_emissions(
                ds, "methyl-chloride"
            ).flatten(),
            "bio_methyl_iodide": interpolate_biogenic_emissions(
                ds, "methyl-iodide"
            ).flatten(),
            "bio_other_aldehydes": interpolate_biogenic_emissions(
                ds, "other-aldehydes"
            ).flatten(),
            "bio_other_ketones": interpolate_biogenic_emissions(
                ds, "other-ketones"
            ).flatten(),
            "bio_other_monoterpenes": interpolate_biogenic_emissions(
                ds, "other-monoterpenes"
            ).flatten(),
            "bio_pinene_a": interpolate_biogenic_emissions(
                ds, "pinene-a"
            ).flatten(),
            "bio_pinene_b": interpolate_biogenic_emissions(
                ds, "pinene-b"
            ).flatten(),
            "bio_propane": interpolate_biogenic_emissions(
                ds, "propane"
            ).flatten(),
            "bio_propene": interpolate_biogenic_emissions(
                ds, "propene"
            ).flatten(),
            "bio_sesquiterpenes": interpolate_biogenic_emissions(
                ds, "sesquiterpenes"
            ).flatten(),
            "bio_toluene": interpolate_biogenic_emissions(
                ds, "toluene"
            ).flatten(),
            "bio_ch2br2": interpolate_biogenic_emissions(
                ds, "CH2Br2"
            ).flatten(),
            "bio_ch3i": interpolate_biogenic_emissions(ds, "CH3I").flatten(),
            "bio_chbr3": interpolate_biogenic_emissions(ds, "CHBr3").flatten(),
            "bio_dms": interpolate_biogenic_emissions(ds, "DMS").flatten(),
        }
    ).set_index(["time", "level"])



from importlib.metadata import version

# Version string for the SOSAA simulation
sosaa_version = version(__package__)
# Pretty-printed version string for the SOSAA GUI
sosaa_version_pretty = f"SOSAA Model {sosaa_version} GUI"
